# --------------------------------------------------------------------

cdef extern from * nogil:

    ctypedef enum PetscDMStagStencilType"DMStagStencilType":
        DMSTAG_STENCIL_STAR
        DMSTAG_STENCIL_BOX
        DMSTAG_STENCIL_NONE

    ctypedef enum PetscDMStagStencilLocation"DMStagStencilLocation":
        DMSTAG_NULL_LOCATION
        DMSTAG_BACK_DOWN_LEFT
        DMSTAG_BACK_DOWN
        DMSTAG_BACK_DOWN_RIGHT
        DMSTAG_BACK_LEFT
        DMSTAG_BACK
        DMSTAG_BACK_RIGHT
        DMSTAG_BACK_UP_LEFT
        DMSTAG_BACK_UP
        DMSTAG_BACK_UP_RIGHT
        DMSTAG_DOWN_LEFT
        DMSTAG_DOWN
        DMSTAG_DOWN_RIGHT
        DMSTAG_LEFT
        DMSTAG_ELEMENT
        DMSTAG_RIGHT
        DMSTAG_UP_LEFT
        DMSTAG_UP
        DMSTAG_UP_RIGHT
        DMSTAG_FRONT_DOWN_LEFT
        DMSTAG_FRONT_DOWN
        DMSTAG_FRONT_DOWN_RIGHT
        DMSTAG_FRONT_LEFT
        DMSTAG_FRONT
        DMSTAG_FRONT_RIGHT
        DMSTAG_FRONT_UP_LEFT
        DMSTAG_FRONT_UP
        DMSTAG_FRONT_UP_RIGHT

    PetscErrorCode DMStagCreate1d(MPI_Comm, PetscDMBoundaryType, PetscInt, PetscInt, PetscInt, PetscDMStagStencilType, PetscInt, const PetscInt[], PetscDM*)
    PetscErrorCode DMStagCreate2d(MPI_Comm, PetscDMBoundaryType, PetscDMBoundaryType, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscDMStagStencilType, PetscInt, const PetscInt[], const PetscInt[], PetscDM*)
    PetscErrorCode DMStagCreate3d(MPI_Comm, PetscDMBoundaryType, PetscDMBoundaryType, PetscDMBoundaryType, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt, PetscDMStagStencilType, PetscInt, const PetscInt[], const PetscInt[], const PetscInt[], PetscDM*)

    PetscErrorCode DMStagGetCorners(PetscDM, PetscInt*, PetscInt*, PetscInt*, PetscInt*, PetscInt*, PetscInt*, PetscInt*, PetscInt*, PetscInt*)
    PetscErrorCode DMStagGetGhostCorners(PetscDM, PetscInt*, PetscInt*, PetscInt*, PetscInt*, PetscInt*, PetscInt*)
    PetscErrorCode DMStagGetLocalSizes(PetscDM, PetscInt*, PetscInt*, PetscInt*)
    PetscErrorCode DMStagGetEntriesPerElement(PetscDM, PetscInt*)

    PetscErrorCode DMStagGetDOF(PetscDM, PetscInt*, PetscInt*, PetscInt*, PetscInt*)
    PetscErrorCode DMStagGetNumRanks(PetscDM, PetscInt*, PetscInt*, PetscInt*)
    PetscErrorCode DMStagGetGlobalSizes(PetscDM, PetscInt*, PetscInt*, PetscInt*)
    PetscErrorCode DMStagGetBoundaryTypes(PetscDM, PetscDMBoundaryType*, PetscDMBoundaryType*, PetscDMBoundaryType*)
    PetscErrorCode DMStagGetStencilWidth(PetscDM, PetscInt*)
    PetscErrorCode DMStagGetStencilType(PetscDM, PetscDMStagStencilType*)
    PetscErrorCode DMStagGetOwnershipRanges(PetscDM, const PetscInt*[], const PetscInt*[], const PetscInt*[])

    PetscErrorCode DMStagSetDOF(PetscDM, PetscInt, PetscInt, PetscInt, PetscInt)
    PetscErrorCode DMStagSetNumRanks(PetscDM, PetscInt, PetscInt, PetscInt)
    PetscErrorCode DMStagSetGlobalSizes(PetscDM, PetscInt, PetscInt, PetscInt)
    PetscErrorCode DMStagSetBoundaryTypes(PetscDM, PetscDMBoundaryType, PetscDMBoundaryType, PetscDMBoundaryType)
    PetscErrorCode DMStagSetStencilWidth(PetscDM, PetscInt)
    PetscErrorCode DMStagSetStencilType(PetscDM, PetscDMStagStencilType)
    PetscErrorCode DMStagSetOwnershipRanges(PetscDM, const PetscInt[], const PetscInt[], const PetscInt[])

    PetscErrorCode DMStagGetLocationSlot(PetscDM, PetscDMStagStencilLocation, PetscInt, PetscInt*)
    PetscErrorCode DMStagGetLocationDOF(PetscDM, PetscDMStagStencilLocation, PetscInt*)
    PetscErrorCode DMStagGetProductCoordinateLocationSlot(PetscDM, PetscDMStagStencilLocation, PetscInt*)

    PetscErrorCode DMStagGetIsFirstRank(PetscDM, PetscBool*, PetscBool*, PetscBool*)
    PetscErrorCode DMStagGetIsLastRank(PetscDM, PetscBool*, PetscBool*, PetscBool*)

    PetscErrorCode DMStagSetUniformCoordinatesExplicit(PetscDM, PetscReal, PetscReal, PetscReal, PetscReal, PetscReal, PetscReal)
    PetscErrorCode DMStagSetUniformCoordinatesProduct(PetscDM, PetscReal, PetscReal, PetscReal, PetscReal, PetscReal, PetscReal)
    PetscErrorCode DMStagSetCoordinateDMType(PetscDM, PetscDMType)
    PetscErrorCode DMStagSetUniformCoordinates(PetscDM, PetscReal, PetscReal, PetscReal, PetscReal, PetscReal, PetscReal)

    PetscErrorCode DMStagCreateCompatibleDMStag(PetscDM, PetscInt, PetscInt, PetscInt, PetscInt, PetscDM*)
    PetscErrorCode DMStagVecSplitToDMDA(PetscDM, PetscVec, PetscDMStagStencilLocation, PetscInt, PetscDM*, PetscVec*)
    PetscErrorCode DMStagMigrateVec(PetscDM, PetscVec, PetscDM, PetscVec)

# --------------------------------------------------------------------

cdef inline PetscDMStagStencilType asStagStencil(object stencil) \
    except <PetscDMStagStencilType>(-1):
    if isinstance(stencil, str):
        if   stencil == "star":  return DMSTAG_STENCIL_STAR
        elif stencil == "box":   return DMSTAG_STENCIL_BOX
        elif stencil == "none":  return DMSTAG_STENCIL_NONE
        else: raise ValueError("unknown stencil type: %s" % stencil)
    return stencil

cdef inline object toStagStencil(PetscDMStagStencilType stype):
    if   stype == DMSTAG_STENCIL_STAR:  return "star"
    elif stype == DMSTAG_STENCIL_BOX:   return "box"
    elif stype == DMSTAG_STENCIL_NONE:  return "none"

cdef inline PetscDMStagStencilLocation asStagStencilLocation(object stencil_location) \
    except <PetscDMStagStencilLocation>(-1):
    if isinstance(stencil_location, str):
        if   stencil_location == "null":                return DMSTAG_NULL_LOCATION
        elif stencil_location == "back_down_left":      return DMSTAG_BACK_DOWN_LEFT
        elif stencil_location == "back_down":           return DMSTAG_BACK_DOWN
        elif stencil_location == "back_down_right":     return DMSTAG_BACK_DOWN_RIGHT
        elif stencil_location == "back_left":           return DMSTAG_BACK_LEFT
        elif stencil_location == "back":                return DMSTAG_BACK
        elif stencil_location == "back_right":          return DMSTAG_BACK_RIGHT
        elif stencil_location == "back_up_left":        return DMSTAG_BACK_UP_LEFT
        elif stencil_location == "back_up":             return DMSTAG_BACK_UP
        elif stencil_location == "back_up_right":       return DMSTAG_BACK_UP_RIGHT
        elif stencil_location == "down_left":           return DMSTAG_DOWN_LEFT
        elif stencil_location == "down":                return DMSTAG_DOWN
        elif stencil_location == "down_right":          return DMSTAG_DOWN_RIGHT
        elif stencil_location == "left":                return DMSTAG_LEFT
        elif stencil_location == "element":             return DMSTAG_ELEMENT
        elif stencil_location == "right":               return DMSTAG_RIGHT
        elif stencil_location == "up_left":             return DMSTAG_UP_LEFT
        elif stencil_location == "up":                  return DMSTAG_UP
        elif stencil_location == "up_right":            return DMSTAG_UP_RIGHT
        elif stencil_location == "front_down_left":     return DMSTAG_FRONT_DOWN_LEFT
        elif stencil_location == "front_down":          return DMSTAG_FRONT_DOWN
        elif stencil_location == "front_down_right":    return DMSTAG_FRONT_DOWN_RIGHT
        elif stencil_location == "front_left":          return DMSTAG_FRONT_LEFT
        elif stencil_location == "front":               return DMSTAG_FRONT
        elif stencil_location == "front_right":         return DMSTAG_FRONT_RIGHT
        elif stencil_location == "front_up_left":       return DMSTAG_FRONT_UP_LEFT
        elif stencil_location == "front_up":            return DMSTAG_FRONT_UP
        elif stencil_location == "front_up_right":      return DMSTAG_FRONT_UP_RIGHT
        else: raise ValueError("unknown stencil location type: %s" % stencil_location)
    return stencil_location


cdef inline PetscInt asStagDims(dims,
                                PetscInt *_M,
                                PetscInt *_N,
                                PetscInt *_P) except? -1:
    cdef PetscInt dim = PETSC_DECIDE
    cdef object M=None, N=None, P=None
    dims = tuple(dims)
    dim = <PetscInt>len(dims)
    if   dim == 0: pass
    elif dim == 1: M, = dims
    elif dim == 2: M, N = dims
    elif dim == 3: M, N, P = dims
    if dim >= 1: _M[0] = asInt(M)
    if dim >= 2: _N[0] = asInt(N)
    if dim >= 3: _P[0] = asInt(P)
    return dim

cdef inline tuple toStagDims(PetscInt dim,
                             PetscInt M,
                             PetscInt N,
                             PetscInt P):
    if   dim == 0: return ()
    elif dim == 1: return (toInt(M),)
    elif dim == 2: return (toInt(M), toInt(N))
    elif dim == 3: return (toInt(M), toInt(N), toInt(P))

cdef inline PetscInt asDofs(dofs,
                            PetscInt *_dof0,
                            PetscInt *_dof1,
                            PetscInt *_dof2,
                            PetscInt *_dof3) except? -1:
    cdef PetscInt ndofs = PETSC_DECIDE
    cdef object dof0=None, dof1=None, dof2=None, dof3=None
    dofs = tuple(dofs)
    ndofs = <PetscInt>len(dofs)
    if ndofs == 2:   dof0, dof1 = dofs
    elif ndofs == 3: dof0, dof1, dof2 = dofs
    elif ndofs == 4: dof0, dof1, dof2, dof3 = dofs
    if ndofs >= 2: _dof0[0] = asInt(dof0)
    if ndofs >= 2: _dof1[0] = asInt(dof1)
    if ndofs >= 3: _dof2[0] = asInt(dof2)
    if ndofs >= 4: _dof3[0] = asInt(dof3)
    return ndofs

cdef inline tuple toDofs(PetscInt ndofs,
                         PetscInt dof0,
                         PetscInt dof1,
                         PetscInt dof2,
                         PetscInt dof3):
    if ndofs == 2: return (toInt(dof0), toInt(dof1))
    elif ndofs == 3: return (toInt(dof0), toInt(dof1), toInt(dof2))
    elif ndofs == 4: return (toInt(dof0), toInt(dof1), toInt(dof2), toInt(dof3))

cdef inline tuple asStagOwnershipRanges(object ownership_ranges,
                                        PetscInt dim,
                                        PetscInt *m, PetscInt *n, PetscInt *p,
                                        PetscInt **_x,
                                        PetscInt **_y,
                                        PetscInt **_z):
    cdef object ranges = list(ownership_ranges)
    cdef PetscInt rdim = <PetscInt>len(ranges)
    cdef PetscInt nlx=0, nly=0, nlz=0
    if dim == PETSC_DECIDE: dim = rdim
    elif dim != rdim: raise ValueError(
        "number of dimensions %d and number ownership ranges %d" %
        (toInt(dim), toInt(rdim)))
    if dim >= 1:
        ranges[0] = iarray_i(ranges[0], &nlx, _x)
        if m[0] == PETSC_DECIDE: m[0] = nlx
        elif m[0] != nlx: raise ValueError(
            "ownership range size %d and number or processors %d" %
            (toInt(nlx), toInt(m[0])))
    if dim >= 2:
        ranges[1] = iarray_i(ranges[1], &nly, _y)
        if n[0] == PETSC_DECIDE: n[0] = nly
        elif n[0] != nly: raise ValueError(
            "ownership range size %d and number or processors %d" %
            (toInt(nly), toInt(n[0])))
    if dim >= 3:
        ranges[2] = iarray_i(ranges[2], &nlz, _z)
        if p[0] == PETSC_DECIDE: p[0] = nlz
        elif p[0] != nlz: raise ValueError(
            "ownership range size %d and number or processors %d" %
            (toInt(nlz), toInt(p[0])))
    return tuple(ranges)


cdef inline tuple toStagOwnershipRanges(PetscInt dim,
                                        PetscInt m, PetscInt n, PetscInt p,
                                        const PetscInt *lx,
                                        const PetscInt *ly,
                                        const PetscInt *lz):
    # Returns tuple of arrays containing ownership ranges as Python arrays
    ranges = [array_i(m, lx)]
    if dim > 1:
        ranges.append(array_i(n, ly))
    if dim > 2:
        ranges.append(array_i(p, lz))
    return tuple(ranges)

cdef inline object toStagBoundary(PetscDMBoundaryType btype):
    if   btype == DM_BOUNDARY_NONE:       return "none"
    elif btype == DM_BOUNDARY_PERIODIC:   return "periodic"
    elif btype == DM_BOUNDARY_GHOSTED:    return "ghosted"

cdef inline tuple toStagBoundaryTypes(PetscInt dim, PetscDMBoundaryType btx, PetscDMBoundaryType bty, PetscDMBoundaryType btz):
    if dim == 1: return (toStagBoundary(btx),)
    if dim == 2: return (toStagBoundary(btx), toStagBoundary(bty))
    if dim == 3: return (toStagBoundary(btx), toStagBoundary(bty), toStagBoundary(btz))

# --------------------------------------------------------------------



from petsc4py.typing import Scalar
from petsc4py.PETSc import Mat
from petsc4py.PETSc import Vec
from petsc4py.PETSc import IS
from petsc4py.PETSc import InsertMode
from petsc4py.PETSc import NormType
from petsc4py.PETSc import Viewer


# A template class with the Python methods supported by MATPYTHON


class MatPythonProtocol:
    def mult(self, A: Mat, x: Vec, y: Vec) -> None:
        """Matrix vector multiplication: y = A @ x."""
        ...

    def multAdd(self, A: Mat, x: Vec, y: Vec, z: Vec) -> None:
        """Matrix vector multiplication: z = A @ x + y."""
        ...

    def multTranspose(self, A: Mat, x: Vec, y: Vec) -> None:
        """Transposed matrix vector multiplication: y = A^T @ x."""
        ...

    def multTransposeAdd(self, A: Mat, x: Vec, y: Vec, z: Vec) -> None:
        """Transposed matrix vector multiplication: z = A^T @ x + y."""
        ...

    def multHermitian(self, A: Mat, x: Vec, y: Vec) -> None:
        """Hermitian matrix vector multiplication: y = A^H @ x."""
        ...

    def multHermitianAdd(self, A: Mat, x: Vec, y: Vec, z: Vec) -> None:
        """Hermitian matrix vector multiplication: z = A^H @ x + y."""
        ...

    def view(self, A: Mat, viewer: Viewer) -> None:
        """View the matrix."""
        ...

    def setFromOptions(self, A: Mat) -> None:
        """Process command line for customization."""
        ...

    def multDiagonalBlock(self, A: Mat, x: Vec, y: Vec) -> None:
        """Perform the on-process matrix vector multiplication."""
        ...

    def createVecs(self, A: Mat) -> tuple[Vec, Vec]:
        """Return tuple of vectors (x,y) suitable for A @ x = y."""
        ...

    def scale(self, A: Mat, s: Scalar) -> None:
        """Scale the matrix by a scalar."""
        ...

    def shift(self, A: Mat, s: Scalar) -> None:
        """Shift the matrix by a scalar."""
        ...

    def createSubMatrix(self, A: Mat, r: IS, c: IS, out: Mat) -> Mat:
        """Return the submatrix corresponding to r rows and c columns.

        Matrix out must be reused if not None.

        """
        ...

    def zeroRowsColumns(self, A: Mat, r: IS, diag: Scalar, x: Vec, b: Vec) -> None:
        """Zero rows and columns of the matrix corresponding to the index set r.

        Insert diag on the diagonal and modify vectors x and b accordingly if not None.

        """
        ...

    def getDiagonal(self, A: Mat, d: Vec) -> None:
        """Compute the diagonal of the matrix: d = diag(A)."""
        ...

    def setDiagonal(self, A: Mat, d: Vec, im: InsertMode) -> None:
        """Set the diagonal of the matrix."""
        ...

    def missingDiagonal(self, A: Mat, d: Vec, im: InsertMode) -> tuple[bool, int]:
        """Return a flag indicating if the matrix is missing a diagonal entry and the location."""
        ...

    def diagonalScale(self, A: Mat, L: Vec, R: Vec) -> None:
        """Perform left and right diagonal scaling if vectors are not None.

        A = diag(L)@A@diag(R).

        """
        ...

    def getDiagonalBlock(self, A: Mat) -> Mat:
        """Return the on-process matrix."""
        ...

    def setUp(self, A: Mat) -> None:
        """Perform the required setup."""
        ...

    def duplicate(self, A: Mat, op: Mat.DuplicateOption) -> Mat:
        """Duplicate the matrix."""
        ...

    def copy(self, A: Mat, B: Mat, op: Mat.Structure) -> None:
        """Copy the matrix: B = A."""
        ...

    def productSetFromOptions(
        self, A: Mat, prodtype: str, X: Mat, Y: Mat, Z: Mat
    ) -> bool:
        """The boolean flag indicating if the matrix supports prodtype."""
        ...

    def productSymbolic(
        self, A: Mat, product: Mat, producttype: str, X: Mat, Y: Mat, Z: Mat
    ) -> None:
        """Perform the symbolic stage of the requested matrix product."""
        ...

    def productNumeric(
        self, A: Mat, product: Mat, producttype: str, X: Mat, Y: Mat, Z: Mat
    ) -> None:
        """Perform the numeric stage of the requested matrix product."""
        ...

    def zeroEntries(self, A: Mat) -> None:
        """Set the matrix to zero."""
        ...

    def norm(self, A: Mat, normtype: NormType) -> float:
        """Compute the norm of the matrix."""
        ...

    def solve(self, A: Mat, y: Vec, x: Vec) -> None:
        """Solve the equation: x = inv(A) y."""
        ...

    def solveAdd(self, A: Mat, y: Vec, z: Vec, x: Vec) -> None:
        """Solve the equation: x = inv(A) y + z."""
        ...

    def solveTranspose(self, A: Mat, y: Vec, x: Vec) -> None:
        """Solve the equation: x = inv(A)^T y."""
        ...

    def solveTransposeAdd(self, A: Mat, y: Vec, z: Vec, x: Vec) -> None:
        """Solve the equation: x = inv(A)^T y + z."""
        ...

    def SOR(
        self,
        A: Mat,
        b: Vec,
        omega: float,
        sortype: Mat.SORType,
        shift: float,
        its: int,
        lits: int,
        x: Vec,
    ) -> None:
        """Perform SOR iterations."""
        ...

    def conjugate(self, A: Mat) -> None:
        """Perform the conjugation of the matrix: A = conj(A)."""
        ...

    def imagPart(self, A: Mat) -> None:
        """Set real part to zero. A = imag(A)."""
        ...

    def realPart(self, A: Mat) -> None:
        """Set imaginary part to zero. A = real(A)."""
        ...



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Typing support."""

from __future__ import annotations  # novermin
from typing import (  # novermin
    Callable,
    Sequence,
    Literal,
)
from numpy.typing import (
    NDArray,
)
from .PETSc import (
    InsertMode,
    ScatterMode,
    NormType,
    Object,
    Vec,
    Mat,
    NullSpace,
    KSP,
    SNES,
    TS,
    TAO,
    TAOLineSearch,
    DM,
)

__all__ = [
    'Scalar',
    'ArrayBool',
    'ArrayInt',
    'ArrayReal',
    'ArrayComplex',
    'ArrayScalar',
    'DimsSpec',
    'AccessModeSpec',
    'InsertModeSpec',
    'ScatterModeSpec',
    'LayoutSizeSpec',
    'NormTypeSpec',
    'PetscOptionsHandlerFunction',
    'MatAssemblySpec',
    'MatSizeSpec',
    'MatBlockSizeSpec',
    'CSRIndicesSpec',
    'CSRSpec',
    'NNZSpec',
    'MatNullFunction',
    'DMCoarsenHookFunction',
    'DMRestrictHookFunction',
    'KSPRHSFunction',
    'KSPOperatorsFunction',
    'KSPConvergenceTestFunction',
    'KSPMonitorFunction',
    'KSPPreSolveFunction',
    'KSPPostSolveFunction',
    'TSRHSFunction',
    'TSRHSJacobian',
    'TSRHSJacobianP',
    'TSIFunction',
    'TSIJacobian',
    'TSIJacobianP',
    'TSI2Function',
    'TSI2Jacobian',
    'TSI2JacobianP',
    'TSMonitorFunction',
    'TSPreStepFunction',
    'TSPostStepFunction',
    'TSIndicatorFunction',
    'TSPostEventFunction',
    'TSPreStepFunction',
    'TSPostStepFunction',
    'TAOObjectiveFunction',
    'TAOGradientFunction',
    'TAOObjectiveGradientFunction',
    'TAOHessianFunction',
    'TAOUpdateFunction',
    'TAOMonitorFunction',
    'TAOConvergedFunction',
    'TAOJacobianFunction',
    'TAOResidualFunction',
    'TAOJacobianResidualFunction',
    'TAOVariableBoundsFunction',
    'TAOConstraintsFunction',
    'TAOConstraintsJacobianFunction',
    'TAOLSObjectiveFunction',
    'TAOLSGradientFunction',
    'TAOLSObjectiveGradientFunction',
]

# --- Sys ---

Scalar = float | complex
"""Scalar type.

Scalars can be either `float` or `complex` (but not both) depending on how
PETSc was configured (``./configure --with-scalar-type=real|complex``).

"""

ArrayBool = NDArray[bool]
"""Array of `bool`."""

ArrayInt = NDArray[int]
"""Array of `int`."""

ArrayReal = NDArray[float]
"""Array of `float`."""

ArrayComplex = NDArray[complex]
"""Array of `complex`."""

ArrayScalar = NDArray[Scalar]
"""Array of `Scalar` numbers."""

DimsSpec = tuple[int, ...]
"""Dimensions specification.

   N-tuples indicates N-dimensional grid sizes.

"""

AccessModeSpec = Literal['rw', 'r', 'w'] | None
"""Access mode specification.

   Possible values are:
     - ``'rw'`` Read-Write mode.
     - ``'r'`` Read-only mode.
     - ``'w'`` Write-only mode.
     - `None` as ``'rw'``.

"""

InsertModeSpec = InsertMode | bool | None
"""Insertion mode specification.

   Possible values are:
     - `InsertMode.ADD_VALUES` Add new value to existing one.
     - `InsertMode.INSERT_VALUES` Replace existing entry with new value.
     - `None` as `InsertMode.INSERT_VALUES`.
     - `False` as `InsertMode.INSERT_VALUES`.
     - `True` as `InsertMode.ADD_VALUES`.

   See Also
   --------
   InsertMode

"""

ScatterModeSpec = ScatterMode | bool | str | None
"""Scatter mode specification.

   Possible values are:
     - `ScatterMode.FORWARD` Forward mode.
     - `ScatterMode.REVERSE` Reverse mode.
     - `None` as `ScatterMode.FORWARD`.
     - `False` as `ScatterMode.FORWARD`.
     - `True` as `ScatterMode.REVERSE`.
     - ``'forward'`` as `ScatterMode.FORWARD`.
     - ``'reverse'`` as `ScatterMode.REVERSE`.

   See Also
   --------
   ScatterMode

"""

LayoutSizeSpec = int | tuple[int, int]
"""`int` or 2-`tuple` of `int` describing the layout sizes.

   A single `int` indicates global size.
   A `tuple` of `int` indicates ``(local_size, global_size)``.

   See Also
   --------
   Sys.splitOwnership

"""

NormTypeSpec = NormType | None
"""Norm type specification.

    Possible values include:

    - `NormType.NORM_1` The 1-norm: Σₙ abs(xₙ) for vectors, maxₙ (Σᵢ abs(xₙᵢ)) for matrices.
    - `NormType.NORM_2` The 2-norm: √(Σₙ xₙ²) for vectors, largest singular values for matrices.
    - `NormType.NORM_INFINITY` The ∞-norm: maxₙ abs(xₙ) for vectors, maxᵢ (Σₙ abs(xₙᵢ)) for matrices.
    - `NormType.NORM_FROBENIUS` The Frobenius norm: same as 2-norm for vectors, √(Σₙᵢ xₙᵢ²) for matrices.
    - `NormType.NORM_1_AND_2` Compute both `NormType.NORM_1` and `NormType.NORM_2`.
    - `None` as `NormType.NORM_2` for vectors, `NormType.NORM_FROBENIUS` for matrices.

    See Also
    --------
    PETSc.NormType, petsc.NormType

"""

# --- PetscObject ---

PetscOptionsHandlerFunction = Callable[[Object], None]
"""Callback for processing extra options."""

# --- Mat ---

MatAssemblySpec = Mat.AssemblyType | bool | None
"""Matrix assembly specification.

   Possible values are:
     - `Mat.AssemblyType.FINAL`
     - `Mat.AssemblyType.FLUSH`
     - `None` as `Mat.AssemblyType.FINAL`
     - `False` as `Mat.AssemblyType.FINAL`
     - `True` as `Mat.AssemblyType.FLUSH`

   See Also
   --------
   petsc.MatAssemblyType

"""

MatSizeSpec = int | tuple[int, int] | tuple[tuple[int, int], tuple[int, int]]
"""`int` or (nested) `tuple` of `int` describing the matrix sizes.

   If `int` then rows = columns.
   A single `tuple` of `int` indicates ``(rows, columns)``.
   A nested `tuple` of `int` indicates ``((local_rows, rows), (local_columns, columns))``.

   See Also
   --------
   Sys.splitOwnership

"""

MatBlockSizeSpec = int | tuple[int, int]
"""The row and column block sizes.

   If a single `int` is provided then rows and columns share the same block size.

"""

CSRIndicesSpec = tuple[Sequence[int], Sequence[int]]
"""CSR indices format specification.

   A 2-tuple carrying the ``(row_start, col_indices)`` information.

"""

CSRSpec = tuple[Sequence[int], Sequence[int], Sequence[Scalar]]
"""CSR format specification.

   A 3-tuple carrying the ``(row_start, col_indices, values)`` information.

"""

NNZSpec = int | Sequence[int] | tuple[Sequence[int], Sequence[int]]
"""Nonzero pattern specification.

   A single `int` corresponds to fixed number of non-zeros per row.
   A `Sequence` of `int` indicates different non-zeros per row.
   If a 2-`tuple` is used, the elements of the tuple corresponds
   to the on-process and off-process parts of the matrix.

   See Also
   --------
   petsc.MatSeqAIJSetPreallocation, petsc.MatMPIAIJSetPreallocation

"""

# --- MatNullSpace ---

MatNullFunction = Callable[[NullSpace, Vec], None]
"""`PETSc.NullSpace` callback."""

# --- DM ---

DMCoarsenHookFunction = Callable[[DM, DM], None]
"""`PETSc.DM` coarsening hook callback."""

DMRestrictHookFunction = Callable[[DM, Mat, Vec, Mat, DM], None]
"""`PETSc.DM` restriction hook callback."""

# --- KSP ---

KSPRHSFunction = Callable[[KSP, Vec], None]
"""`PETSc.KSP` right-hand side function callback."""

KSPOperatorsFunction = Callable[[KSP, Mat, Mat], None]
"""`PETSc.KSP` operators function callback."""

KSPConvergenceTestFunction = Callable[[KSP, int, float], KSP.ConvergedReason]
"""`PETSc.KSP` convergence test callback."""

KSPMonitorFunction = Callable[[KSP, int, float], None]
"""`PETSc.KSP` monitor callback."""

KSPPreSolveFunction = Callable[[KSP, Vec, Vec], None]
"""`PETSc.KSP` pre solve callback."""

KSPPostSolveFunction = Callable[[KSP, Vec, Vec], None]
"""`PETSc.KSP` post solve callback."""

# --- SNES ---

SNESMonitorFunction = Callable[[SNES, int, float], None]
"""`SNES` monitor callback."""

SNESObjFunction = Callable[[SNES, Vec], None]
"""`SNES` objective function callback."""

SNESFunction = Callable[[SNES, Vec, Vec], None]
"""`SNES` residual function callback."""

SNESJacobianFunction = Callable[[SNES, Vec, Mat, Mat], None]
"""`SNES` Jacobian callback."""

SNESGuessFunction = Callable[[SNES, Vec], None]
"""`SNES` initial guess callback."""

SNESUpdateFunction = Callable[[SNES, int], None]
"""`SNES` step update callback."""

SNESLSPreFunction = Callable[[Vec, Vec], None]
"""`SNES` linesearch pre-check update callback."""

SNESNGSFunction = Callable[[SNES, Vec, Vec], None]
"""`SNES` nonlinear Gauss-Seidel callback."""

SNESConvergedFunction = Callable[
    [SNES, int, tuple[float, float, float]], SNES.ConvergedReason
]
"""`SNES` convergence test callback."""

# --- TS ---

TSRHSFunction = Callable[[TS, float, Vec, Vec], None]
"""`TS` right-hand side function callback."""

TSRHSJacobian = Callable[[TS, float, Vec, Mat, Mat], None]
"""`TS` right-hand side Jacobian callback."""

TSRHSJacobianP = Callable[[TS, float, Vec, Mat], None]
"""`TS` right-hand side parameter Jacobian callback."""

TSIFunction = Callable[[TS, float, Vec, Vec, Vec], None]
"""`TS` implicit function callback."""

TSIJacobian = Callable[[TS, float, Vec, Vec, float, Mat, Mat], None]
"""`TS` implicit Jacobian callback."""

TSIJacobianP = Callable[[TS, float, Vec, Vec, float, Mat], None]
"""`TS` implicit parameter Jacobian callback."""

TSI2Function = Callable[[TS, float, Vec, Vec, Vec, Vec], None]
"""`TS` implicit 2nd order function callback."""

TSI2Jacobian = Callable[[TS, float, Vec, Vec, Vec, float, float, Mat, Mat], None]
"""`TS` implicit 2nd order Jacobian callback."""

TSI2JacobianP = Callable[[TS, float, Vec, Vec, Vec, float, float, Mat], None]
"""`TS` implicit 2nd order parameter Jacobian callback."""

TSMonitorFunction = Callable[[TS, int, float, Vec], None]
"""`TS` monitor callback."""

TSPreStepFunction = Callable[[TS], None]
"""`TS` pre-step callback."""

TSPostStepFunction = Callable[[TS], None]
"""`TS` post-step callback."""

TSIndicatorFunction = Callable[[TS, float, Vec, NDArray[float]], None]
"""`TS` event indicator callback."""

TSPostEventFunction = Callable[[TS, NDArray[int], float, Vec, bool], None]
"""`TS` post-event callback."""

# --- TAO ---

TAOObjectiveFunction = Callable[[TAO, Vec], float]
"""`TAO` objective function callback."""

TAOGradientFunction = Callable[[TAO, Vec, Vec], None]
"""`TAO` objective gradient callback."""

TAOObjectiveGradientFunction = Callable[[TAO, Vec, Vec], float]
"""`TAO` objective function and gradient callback."""

TAOHessianFunction = Callable[[TAO, Vec, Mat, Mat], None]
"""`TAO` objective Hessian callback."""

TAOUpdateFunction = Callable[[TAO, int], None]
"""`TAO` update callback."""

TAOMonitorFunction = Callable[[TAO], None]
"""`TAO` monitor callback."""

TAOConvergedFunction = Callable[[TAO], None]
"""`TAO` convergence test callback."""

TAOJacobianFunction = Callable[[TAO, Vec, Mat, Mat], None]
"""`TAO` Jacobian callback."""

TAOResidualFunction = Callable[[TAO, Vec, Vec], None]
"""`TAO` residual callback."""

TAOJacobianResidualFunction = Callable[[TAO, Vec, Mat, Mat], None]
"""`TAO` Jacobian residual callback."""

TAOVariableBoundsFunction = Callable[[TAO, Vec, Vec], None]
"""`TAO` variable bounds callback."""

TAOConstraintsFunction = Callable[[TAO, Vec, Vec], None]
"""`TAO` constraints callback."""

TAOConstraintsJacobianFunction = Callable[[TAO, Vec, Mat, Mat], None]
"""`TAO` constraints Jacobian callback."""

TAOLSObjectiveFunction = Callable[[TAOLineSearch, Vec], float]
"""`TAOLineSearch` objective function callback."""

TAOLSGradientFunction = Callable[[TAOLineSearch, Vec, Vec], None]
"""`TAOLineSearch` objective gradient callback."""

TAOLSObjectiveGradientFunction = Callable[[TAOLineSearch, Vec, Vec], float]
"""`TAOLineSearch` objective function and gradient callback."""



import sys, petsc4py
petsc4py.init(sys.argv)

import numpy as np
from petsc4py import PETSc

class Chwirut(object):

    """
    Finds the nonlinear least-squares solution to the model
    y = exp(-b1*x)/(b2+b3*x)  +  e
    """

    def __init__(self):
        BETA = [0.2, 0.12, 0.08]
        NOBSERVATIONS = 100
        NPARAMETERS = 3

        np.random.seed(456)
        x = np.random.rand(NOBSERVATIONS)
        e = np.random.rand(NOBSERVATIONS)

        y = np.exp(-BETA[0]*x)/(BETA[1] + BETA[2]*x) + e

        self.NOBSERVATIONS = NOBSERVATIONS
        self.NPARAMETERS = NPARAMETERS
        self.x = x
        self.y = y

    def createVecs(self):
        X = PETSc.Vec().create(PETSc.COMM_SELF)
        X.setSizes(self.NPARAMETERS)
        F = PETSc.Vec().create(PETSc.COMM_SELF)
        F.setSizes(self.NOBSERVATIONS)
        return X, F

    def formInitialGuess(self, X):
        X[0] = 0.15
        X[1] = 0.08
        X[2] = 0.05

    def formResidual(self, tao, X, F):
        x, y = self.x, self.y
        b1, b2, b3 = X.array
        F.array = y - np.exp(-b1*x)/(b2 + b3*x)

    def plotSolution(self, X):
        try:
            from matplotlib import pylab
        except ImportError:
            return
        b1, b2, b3 = X.array
        x, y = self.x, self.y
        u = np.linspace(x.min(), x.max(), 100)
        v = np.exp(-b1*u)/(b2+b3*u)
        pylab.plot(x, y, 'ro')
        pylab.plot(u, v, 'b-')
        pylab.show()

OptDB = PETSc.Options()

user = Chwirut()

x, f = user.createVecs()
x.setFromOptions()
f.setFromOptions()

tao = PETSc.TAO().create(PETSc.COMM_SELF)
tao.setType(PETSc.TAO.Type.POUNDERS)
tao.setResidual(user.formResidual, f)
tao.setFromOptions()

user.formInitialGuess(x)
tao.solve(x)

plot = OptDB.getBool('plot', False)
if plot: user.plotSolution(x)

x.destroy()
f.destroy()
tao.destroy()



# --------------------------------------------------------------------

class RandomType(object):
    """The random number generator type."""
    RAND      = S_(PETSCRAND)
    RAND48    = S_(PETSCRAND48)
    SPRNG     = S_(PETSCSPRNG)
    RANDER48  = S_(PETSCRANDER48)
    RANDOM123 = S_(PETSCRANDOM123)

# --------------------------------------------------------------------


cdef class Random(Object):
    """The random number generator object.

    See Also
    --------
    petsc.PetscRandom

    """

    Type = RandomType

    def __cinit__(self) -> None:
        self.obj = <PetscObject*> &self.rnd
        self.rnd = NULL

    def __call__(self) -> Scalar:
        """Generate a scalar random number.

        Not collective.

        See Also
        --------
        petsc.PetscRandomGetValue

        """
        return self.getValue()

    def view(self, Viewer viewer=None) -> None:
        """View a random number generator object.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` instance or `None` for the default viewer.

        See Also
        --------
        petsc.PetscRandomView

        """
        assert self.obj != NULL
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(PetscRandomView(self.rnd, vwr))

    def destroy(self) -> Self:
        """Destroy the random number generator object.

        Collective.

        See Also
        --------
        petsc.PetscRandomDestroy

        """
        CHKERR(PetscRandomDestroy(&self.rnd))
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """Create a random number generator object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        Sys.getDefaultComm, petsc.PetscRandomCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        CHKERR(PetscRandomCreate(ccomm, &self.rnd))
        return self

    def setType(self, rnd_type: Random.Type | str) -> None:
        """Set the type of the random number generator object.

        Collective.

        Parameters
        ----------
        rnd_type
            The type of the generator.

        See Also
        --------
        getType, petsc.PetscRandomSetType

        """
        cdef PetscRandomType cval = NULL
        rnd_type = str2bytes(rnd_type, &cval)
        CHKERR(PetscRandomSetType(self.rnd, cval))

    def getType(self) -> str:
        """Return the type of the random number generator object.

        Not collective.

        See Also
        --------
        setType, petsc.PetscRandomGetType

        """
        cdef PetscRandomType cval = NULL
        CHKERR(PetscRandomGetType(self.rnd, &cval))
        return bytes2str(cval)

    def setFromOptions(self) -> None:
        """Configure the random number generator from the options database.

        Collective.

        See Also
        --------
        petsc_options, petsc.PetscRandomSetFromOptions

        """
        CHKERR(PetscRandomSetFromOptions(self.rnd))

    def getValue(self) -> Scalar:
        """Generate a scalar random number.

        Not collective.

        See Also
        --------
        petsc.PetscRandomGetValue

        """
        cdef PetscScalar sval = 0
        CHKERR(PetscRandomGetValue(self.rnd, &sval))
        return toScalar(sval)

    def getValueReal(self) -> float:
        """Generate a real random number.

        Not collective.

        See Also
        --------
        petsc.PetscRandomGetValueReal

        """
        cdef PetscReal rval = 0
        CHKERR(PetscRandomGetValueReal(self.rnd, &rval))
        return toReal(rval)

    def getSeed(self) -> int:
        """Return the random number generator seed.

        Not collective.

        See Also
        --------
        setSeed, petsc.PetscRandomGetSeed

        """
        cdef PetscInt64 seed = 0
        CHKERR(PetscRandomGetSeed(self.rnd, &seed))
        return seed

    def setSeed(self, seed: int | None = None) -> None:
        """Set the seed of random number generator.

        Not collective.

        Parameters
        ----------
        seed
            The value for the seed. If `None`, it only seeds the generator.

        See Also
        --------
        getSeed, petsc.PetscRandomSetSeed, petsc.PetscRandomSeed

        """
        if seed is not None:
            CHKERR(PetscRandomSetSeed(self.rnd, seed))
        CHKERR(PetscRandomSeed(self.rnd))

    def getInterval(self) -> tuple[Scalar, Scalar]:
        """Return the interval containing the random numbers generated.

        Not collective.

        See Also
        --------
        setInterval, petsc.PetscRandomGetInterval

        """
        cdef PetscScalar sval1 = 0
        cdef PetscScalar sval2 = 1
        CHKERR(PetscRandomGetInterval(self.rnd, &sval1, &sval2))
        return (toScalar(sval1), toScalar(sval2))

    def setInterval(self, interval: tuple[Scalar, Scalar]) -> None:
        """Set the interval of the random number generator.

        Not collective.

        See Also
        --------
        getInterval, petsc.PetscRandomSetInterval

        """
        cdef PetscScalar sval1 = 0
        cdef PetscScalar sval2 = 1
        low, high = interval
        sval1 = asScalar(low)
        sval2 = asScalar(high)
        CHKERR(PetscRandomSetInterval(self.rnd, sval1, sval2))

    #

    property seed:
        """The seed of the random number generator."""
        def __get__(self) -> int:
            return self.getSeed()

        def __set__(self, value: int | None) -> None:
            self.setSeed(value)

    property interval:
        """The interval of the generated random numbers."""
        def __get__(self) -> tuple[Scalar, Scalar]:
            return self.getInterval()

        def __set__(self, value: tuple[Scalar, Scalar]):
            self.setInterval(value)

# --------------------------------------------------------------------

del RandomType

# --------------------------------------------------------------------



#!/usr/bin/env python3

#$ python setup.py build_ext --inplace

from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import cythonize

import numpy
import petsc4py

def configure():
    INCLUDE_DIRS = []
    LIBRARY_DIRS = []
    LIBRARIES    = []

    # PETSc
    import os
    PETSC_DIR  = os.environ['PETSC_DIR']
    PETSC_ARCH = os.environ.get('PETSC_ARCH', '')
    from os.path import join, isdir
    if PETSC_ARCH and isdir(join(PETSC_DIR, PETSC_ARCH)):
        INCLUDE_DIRS += [join(PETSC_DIR, PETSC_ARCH, 'include'),
                         join(PETSC_DIR, 'include')]
        LIBRARY_DIRS += [join(PETSC_DIR, PETSC_ARCH, 'lib')]
    else:
        if PETSC_ARCH: pass # XXX should warn ...
        INCLUDE_DIRS += [join(PETSC_DIR, 'include')]
        LIBRARY_DIRS += [join(PETSC_DIR, 'lib')]
    LIBRARIES += ['petsc']

    # PETSc for Python
    INCLUDE_DIRS += [petsc4py.get_include()]

    # NumPy
    INCLUDE_DIRS += [numpy.get_include()]

    return dict(
        include_dirs=INCLUDE_DIRS + [os.curdir],
        libraries=LIBRARIES,
        library_dirs=LIBRARY_DIRS,
        runtime_library_dirs=LIBRARY_DIRS,
    )

extensions = [
    Extension('Bratu3D',
              sources = ['Bratu3D.pyx',
                         'Bratu3Dimpl.c'],
              depends = ['Bratu3Dimpl.h'],
              **configure()),
]

setup(name = "Bratu3D",
      ext_modules = cythonize(
          extensions, include_path=[petsc4py.get_include()]),
)



# Stiff 3-variable ODE system from chemical reactions,
# due to Robertson (1966),
# problem ROBER in Hairer&Wanner, ODE 2, 1996

import sys
import petsc4py

petsc4py.init(sys.argv)

from petsc4py import PETSc


class Rober:
    n = 3
    comm = PETSc.COMM_SELF

    def evalSolution(self, t, x):
        if t != 0.0:
            raise ValueError('Only for t=0')
        x[:] = [1, 0, 0]
        x.assemble()

    def evalFunction(self, ts, t, x, xdot, f):
        f[:] = [
            xdot[0] + 0.04 * x[0] - 1e4 * x[1] * x[2],
            xdot[1] - 0.04 * x[0] + 1e4 * x[1] * x[2] + 3e7 * x[1] ** 2,
            xdot[2] - 3e7 * x[1] ** 2,
        ]
        f.assemble()

    def evalJacobian(self, ts, t, x, xdot, a, A, B):
        J = B
        J[:, :] = [
            [a + 0.04, -1e4 * x[2], -1e4 * x[1]],
            [-0.04, a + 1e4 * x[2] + 3e7 * 2 * x[1], 1e4 * x[1]],
            [0, -3e7 * 2 * x[1], a],
        ]
        J.assemble()
        if A != B:
            A.assemble()


OptDB = PETSc.Options()
ode = Rober()

J = PETSc.Mat().createDense([ode.n, ode.n], comm=ode.comm)
J.setUp()
x = PETSc.Vec().createSeq(ode.n, comm=ode.comm)
f = x.duplicate()

ts = PETSc.TS().create(comm=ode.comm)
ts.setProblemType(ts.ProblemType.NONLINEAR)
ts.setType(ts.Type.THETA)

ts.setIFunction(ode.evalFunction, f)
ts.setIJacobian(ode.evalJacobian, J)

history = []


def monitor(ts, i, t, x):
    xx = x[:].tolist()
    history.append((i, t, xx))


ts.setMonitor(monitor)

ts.setTime(0.0)
ts.setTimeStep(0.001)
ts.setMaxTime(1e30)
ts.setMaxSteps(100)
ts.setExactFinalTime(PETSc.TS.ExactFinalTime.INTERPOLATE)

ts.setFromOptions()
ode.evalSolution(0.0, x)
ts.solve(x)

del ode, J, x, ts

from matplotlib import pylab
import numpy as np

ii = np.asarray([v[0] for v in history])
tt = np.asarray([v[1] for v in history])
xx = np.asarray([v[2] for v in history])

pylab.suptitle('Rober')
pylab.subplot(2, 2, 1)
pylab.subplots_adjust(wspace=0.3)
pylab.semilogy(
    ii[:-1],
    np.diff(tt),
)
pylab.xlabel('step number')
pylab.ylabel('timestep')

for i in range(3):
    pylab.subplot(2, 2, i + 2)
    pylab.semilogx(tt, xx[:, i], 'rgb'[i])
    pylab.xlabel('t')
    pylab.ylabel('x%d' % i)

pylab.show()



from petsc4py.PETSc import KSP
from petsc4py.PETSc import Vec
from petsc4py.PETSc import Viewer


# A template class with the Python methods supported by KSPPYTHON


class KSPPythonProtocol:
    def solve(self, ksp: KSP, b: Vec, x: Vec) -> None:
        """Solve the linear system with right-hand side b. Return solution in x."""
        ...

    def solveTranspose(self, ksp: KSP, b: Vec, x: Vec) -> None:
        """Solve the transposed linear system with right-hand side b. Return solution in x."""
        ...

    def view(self, ksp: KSP, viewer: Viewer) -> None:
        """View the Krylov solver."""
        ...

    def setFromOptions(self, ksp: KSP) -> None:
        """Process command line for customization."""
        ...

    def setUp(self, ksp: KSP) -> None:
        """Perform the required setup."""
        ...

    def buildSolution(self, ksp: KSP, x: Vec) -> None:
        """Compute the solution vector."""
        ...

    def buildResidual(self, ksp: KSP, t: Vec, r: Vec) -> None:
        """Compute the residual vector, return it in r. t is a scratch working vector."""
        ...

    def reset(self, ksp: KSP) -> None:
        """Reset the Krylov solver."""
        ...



# Stiff scalar valued ODE problem with an exact solution

import sys
import petsc4py

petsc4py.init(sys.argv)

from petsc4py import PETSc
from math import sin, cos, exp


class CE:
    n = 1
    comm = PETSc.COMM_SELF

    def __init__(self, lambda_=1.0):
        self.lambda_ = lambda_

    def evalSolution(self, t, x):
        lam = self.lambda_
        x[0] = lam / (lam * lam + 1) * (lam * cos(t) + sin(t)) - lam * lam / (
            lam * lam + 1
        ) * exp(-lam * t)
        x.assemble()

    def evalFunction(self, ts, t, x, xdot, f):
        lam = self.lambda_
        f[0] = xdot[0] + lam * (x[0] - cos(t))
        f.assemble()

    def evalJacobian(self, ts, t, x, xdot, a, A, B):
        J = B
        lam = self.lambda_
        J[0, 0] = a + lam
        J.assemble()
        if A != B:
            A.assemble()


OptDB = PETSc.Options()

lambda_ = OptDB.getScalar('lambda', 10.0)
ode = CE(lambda_)

J = PETSc.Mat().createDense([ode.n, ode.n], comm=ode.comm)
J.setUp()
x = PETSc.Vec().createSeq(ode.n, comm=ode.comm)
f = x.duplicate()

ts = PETSc.TS().create(comm=ode.comm)
ts.setProblemType(ts.ProblemType.NONLINEAR)
ts.setType(ts.Type.GLLE)

ts.setIFunction(ode.evalFunction, f)
ts.setIJacobian(ode.evalJacobian, J)

ts.setTime(0.0)
ts.setTimeStep(0.001)
ts.setMaxTime(10)
ts.setMaxSteps(10000)
ts.setExactFinalTime(PETSc.TS.ExactFinalTime.INTERPOLATE)


class Monitor:
    def __init__(self, ode):
        self.ode = ode
        self.x = PETSc.Vec().createSeq(ode.n, comm=ode.comm)

    def __call__(self, ts, k, t, x):
        self.ode.evalSolution(t, self.x)
        self.x.axpy(-1, x)
        e = self.x.norm()
        h = ts.getTimeStep()
        PETSc.Sys.Print(
            'step %3d t=%8.2e h=%8.2e error=%8.2e' % (k, t, h, e), comm=self.ode.comm
        )


ts.setMonitor(Monitor(ode))

ts.setFromOptions()
ode.evalSolution(0.0, x)
ts.solve(x)

del ode, J, x, f, ts



# --------------------------------------------------------------------

class DMStagStencilType(object):
    """Stencil types."""
    STAR = DMSTAG_STENCIL_STAR
    BOX  = DMSTAG_STENCIL_BOX
    NONE = DMSTAG_STENCIL_NONE


class DMStagStencilLocation(object):
    """Stencil location types."""
    NULLLOC          = DMSTAG_NULL_LOCATION
    BACK_DOWN_LEFT   = DMSTAG_BACK_DOWN_LEFT
    BACK_DOWN        = DMSTAG_BACK_DOWN
    BACK_DOWN_RIGHT  = DMSTAG_BACK_DOWN_RIGHT
    BACK_LEFT        = DMSTAG_BACK_LEFT
    BACK             = DMSTAG_BACK
    BACK_RIGHT       = DMSTAG_BACK_RIGHT
    BACK_UP_LEFT     = DMSTAG_BACK_UP_LEFT
    BACK_UP          = DMSTAG_BACK_UP
    BACK_UP_RIGHT    = DMSTAG_BACK_UP_RIGHT
    DOWN_LEFT        = DMSTAG_DOWN_LEFT
    DOWN             = DMSTAG_DOWN
    DOWN_RIGHT       = DMSTAG_DOWN_RIGHT
    LEFT             = DMSTAG_LEFT
    ELEMENT          = DMSTAG_ELEMENT
    RIGHT            = DMSTAG_RIGHT
    UP_LEFT          = DMSTAG_UP_LEFT
    UP               = DMSTAG_UP
    UP_RIGHT         = DMSTAG_UP_RIGHT
    FRONT_DOWN_LEFT  = DMSTAG_FRONT_DOWN_LEFT
    FRONT_DOWN       = DMSTAG_FRONT_DOWN
    FRONT_DOWN_RIGHT = DMSTAG_FRONT_DOWN_RIGHT
    FRONT_LEFT       = DMSTAG_FRONT_LEFT
    FRONT            = DMSTAG_FRONT
    FRONT_RIGHT      = DMSTAG_FRONT_RIGHT
    FRONT_UP_LEFT    = DMSTAG_FRONT_UP_LEFT
    FRONT_UP         = DMSTAG_FRONT_UP
    FRONT_UP_RIGHT   = DMSTAG_FRONT_UP_RIGHT

# --------------------------------------------------------------------


cdef class DMStag(DM):
    """A DM object representing a "staggered grid" or a structured cell complex."""

    StencilType       = DMStagStencilType
    StencilLocation   = DMStagStencilLocation

    def create(
        self,
        dim: int,
        dofs: tuple[int, ...] | None = None,
        sizes: tuple[int, ...] | None = None,
        boundary_types: tuple[DM.BoundaryType | int | str | bool, ...] | None = None,
        stencil_type: StencilType | None = None,
        stencil_width: int | None = None,
        proc_sizes: tuple[int, ...] | None = None,
        ownership_ranges: tuple[Sequence[int], ...] | None = None,
        comm: Comm | None = None,
        setUp: bool | None = False) -> Self:
        """Create a DMDA object.

        Collective.

        Creates an object to manage data living on the elements and vertices /
        the elements, faces, and vertices / the elements, faces, edges, and
        vertices of a parallelized regular 1D / 2D / 3D grid.

        Parameters
        ----------
        dim
            The number of dimensions.
        dofs
            The number of degrees of freedom per vertex, element (1D); vertex,
            face, element (2D); or vertex, edge, face, element (3D).
        sizes
            The number of elements in each dimension.
        boundary_types
            The boundary types.
        stencil_type
            The ghost/halo stencil type.
        stencil_width
            The width of the ghost/halo region.
        proc_sizes
            The number of processes in x, y, z dimensions.
        ownership_ranges
            Local x, y, z element counts, of length equal to ``proc_sizes``,
            summing to ``sizes``.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.
        setUp
            Whether to call the setup routine after creating the object.

        See Also
        --------
        petsc.DMStagCreate1d, petsc.DMStagCreate2d, petsc.DMStagCreate3d
        petsc.DMSetUp

        """
        # ndim
        cdef PetscInt ndim = asInt(dim)

        # sizes
        cdef object gsizes = sizes
        cdef PetscInt nsizes=PETSC_DECIDE, M=1, N=1, P=1
        if sizes is not None:
            nsizes = asStagDims(gsizes, &M, &N, &P)
            assert(nsizes==ndim)

        # dofs
        cdef object cdofs = dofs
        cdef PetscInt ndofs=PETSC_DECIDE, dof0=1, dof1=0, dof2=0, dof3=0
        if dofs is not None:
            ndofs = asDofs(cdofs, &dof0, &dof1, &dof2, &dof3)
            assert(ndofs==ndim+1)

        # boundary types
        cdef PetscDMBoundaryType btx = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType bty = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType btz = DM_BOUNDARY_NONE
        asBoundary(boundary_types, &btx, &bty, &btz)

        # stencil
        cdef PetscInt swidth = 0
        if stencil_width is not None:
            swidth = asInt(stencil_width)
        cdef PetscDMStagStencilType stype = DMSTAG_STENCIL_NONE
        if stencil_type is not None:
            stype = asStagStencil(stencil_type)

        # comm
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)

        # proc sizes
        cdef object psizes = proc_sizes
        cdef PetscInt nprocs=PETSC_DECIDE, m=PETSC_DECIDE, n=PETSC_DECIDE, p=PETSC_DECIDE
        if proc_sizes is not None:
            nprocs = asStagDims(psizes, &m, &n, &p)
            assert(nprocs==ndim)

        # ownership ranges
        cdef PetscInt *lx = NULL, *ly = NULL, *lz = NULL
        if ownership_ranges is not None:
            nranges = asStagOwnershipRanges(ownership_ranges, ndim, &m, &n, &p, &lx, &ly, &lz)
            assert(nranges==ndim)

        # create
        cdef PetscDM newda = NULL
        if dim == 1:
            CHKERR(DMStagCreate1d(ccomm, btx, M, dof0, dof1, stype, swidth, lx, &newda))
        if dim == 2:
            CHKERR(DMStagCreate2d(ccomm, btx, bty, M, N, m, n, dof0, dof1, dof2, stype, swidth, lx, ly, &newda))
        if dim == 3:
            CHKERR(DMStagCreate3d(ccomm, btx, bty, btz, M, N, P, m, n, p, dof0, dof1, dof2, dof3, stype, swidth, lx, ly, lz, &newda))
        CHKERR(PetscCLEAR(self.obj)); self.dm = newda
        if setUp:
            CHKERR(DMSetUp(self.dm))
        return self

    # Setters

    def setStencilWidth(self, swidth: int) -> None:
        """Set elementwise stencil width.

        Logically collective.

        The width value is not used when `StencilType.NONE` is specified.

        Parameters
        ----------
        swidth
            Stencil/halo/ghost width in elements.

        See Also
        --------
        petsc.DMStagSetStencilWidth

        """
        cdef PetscInt sw = asInt(swidth)
        CHKERR(DMStagSetStencilWidth(self.dm, sw))

    def setStencilType(self, stenciltype: StencilType | str) -> None:
        """Set elementwise ghost/halo stencil type.

        Logically collective.

        Parameters
        ----------
        stenciltype
            The elementwise ghost stencil type.

        See Also
        --------
        getStencilType, petsc.DMStagSetStencilType

        """
        cdef PetscDMStagStencilType stype = asStagStencil(stenciltype)
        CHKERR(DMStagSetStencilType(self.dm, stype))

    def setBoundaryTypes(
        self,
        boundary_types: tuple[DM.BoundaryType | int | str | bool, ...]) -> None:
        """Set the boundary types.

        Logically collective.

        Parameters
        ----------
        boundary_types
            Boundary types for one/two/three dimensions.

        See Also
        --------
        getBoundaryTypes, petsc.DMStagSetBoundaryTypes

        """
        cdef PetscDMBoundaryType btx = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType bty = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType btz = DM_BOUNDARY_NONE
        asBoundary(boundary_types, &btx, &bty, &btz)
        CHKERR(DMStagSetBoundaryTypes(self.dm, btx, bty, btz))

    def setDof(self, dofs: tuple[int, ...]) -> None:
        """Set DOFs/stratum.

        Logically collective.

        Parameters
        ----------
        dofs
            The number of points per 0-cell (vertex/node), 1-cell (element in
            1D, edge in 2D and 3D), 2-cell (element in 2D, face in 3D), or
            3-cell (element in 3D).

        See Also
        --------
        petsc.DMStagSetDOF

        """
        cdef tuple gdofs = tuple(dofs)
        cdef PetscInt dof0=1, dof1=0, dof2=0, dof3=0
        asDofs(gdofs, &dof0, &dof1, &dof2, &dof3)
        CHKERR(DMStagSetDOF(self.dm, dof0, dof1, dof2, dof3))

    def setGlobalSizes(self, sizes: tuple[int, ...]) -> None:
        """Set global element counts in each dimension.

        Logically collective.

        Parameters
        ----------
        sizes
            Global elementwise size in the one/two/three dimensions.

        See Also
        --------
        petsc.DMStagSetGlobalSizes

        """
        cdef tuple gsizes = tuple(sizes)
        cdef PetscInt M=1, N=1, P=1
        asStagDims(gsizes, &M, &N, &P)
        CHKERR(DMStagSetGlobalSizes(self.dm, M, N, P))

    def setProcSizes(self, sizes: tuple[int, ...]) -> None:
        """Set the number of processes in each dimension in the global process grid.

        Logically collective.

        Parameters
        ----------
        sizes
            Number of processes in one/two/three dimensions.

        See Also
        --------
        petsc.DMStagSetNumRanks

        """
        cdef tuple psizes = tuple(sizes)
        cdef PetscInt m=PETSC_DECIDE, n=PETSC_DECIDE, p=PETSC_DECIDE
        asStagDims(psizes, &m, &n, &p)
        CHKERR(DMStagSetNumRanks(self.dm, m, n, p))

    def setOwnershipRanges(self, ranges: tuple[Sequence[int], ...]) -> None:
        """Set elements per process in each dimension.

        Logically collective.

        Parameters
        ----------
        ranges
            Element counts for each process in one/two/three dimensions.

        See Also
        --------
        getOwnershipRanges, petsc.DMStagSetOwnershipRanges

        """
        cdef PetscInt dim=0, m=PETSC_DECIDE, n=PETSC_DECIDE, p=PETSC_DECIDE
        cdef PetscInt *lx = NULL, *ly = NULL, *lz = NULL
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetNumRanks(self.dm, &m, &n, &p))
        asStagOwnershipRanges(ranges, dim, &m, &n, &p, &lx, &ly, &lz)
        CHKERR(DMStagSetOwnershipRanges(self.dm, lx, ly, lz))

    # Getters

    def getDim(self) -> int:
        """Return the number of dimensions.

        Not collective.

        """
        return self.getDimension()

    def getEntriesPerElement(self) -> int:
        """Return the number of entries per element in the local representation.

        Not collective.

        This is the natural block size for most local operations.

        See Also
        --------
        petsc.DMStagGetEntriesPerElement

        """
        cdef PetscInt epe=0
        CHKERR(DMStagGetEntriesPerElement(self.dm, &epe))
        return toInt(epe)

    def getStencilWidth(self) -> int:
        """Return elementwise stencil width.

        Not collective.

        See Also
        --------
        petsc.DMStagGetStencilWidth

        """
        cdef PetscInt swidth=0
        CHKERR(DMStagGetStencilWidth(self.dm, &swidth))
        return toInt(swidth)

    def getDof(self) -> tuple[int, ...]:
        """Get number of DOFs associated with each stratum of the grid.

        Not collective.

        See Also
        --------
        petsc.DMStagGetDOF

        """
        cdef PetscInt dim=0, dof0=0, dof1=0, dof2=0, dof3=0
        CHKERR(DMStagGetDOF(self.dm, &dof0, &dof1, &dof2, &dof3))
        CHKERR(DMGetDimension(self.dm, &dim))
        return toDofs(dim+1, dof0, dof1, dof2, dof3)

    def getCorners(self) -> tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...]]:
        """Return starting element index, width and number of partial elements.

        Not collective.

        The returned value is calculated excluding ghost points.

        The number of extra partial elements is either ``1`` or ``0``. The
        value is ``1`` on right, top, and front non-periodic domain
        ("physical") boundaries, in the x, y, and z dimensions respectively,
        and otherwise ``0``.

        See Also
        --------
        getGhostCorners, petsc.DMStagGetCorners, petsc.DMGetDimension

        """
        cdef PetscInt dim=0, x=0, y=0, z=0, m=0, n=0, p=0, nExtrax=0, nExtray=0, nExtraz=0
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetCorners(self.dm, &x, &y, &z, &m, &n, &p, &nExtrax, &nExtray, &nExtraz))
        return (asInt(x), asInt(y), asInt(z))[:<Py_ssize_t>dim], (asInt(m), asInt(n), asInt(p))[:<Py_ssize_t>dim], (asInt(nExtrax), asInt(nExtray), asInt(nExtraz))[:<Py_ssize_t>dim]

    def getGhostCorners(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
        """Return starting element index and width of local region.

        Not collective.

        See Also
        --------
        getCorners, petsc.DMStagGetGhostCorners

        """
        cdef PetscInt dim=0, x=0, y=0, z=0, m=0, n=0, p=0
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetGhostCorners(self.dm, &x, &y, &z, &m, &n, &p))
        return (asInt(x), asInt(y), asInt(z))[:<Py_ssize_t>dim], (asInt(m), asInt(n), asInt(p))[:<Py_ssize_t>dim]

    def getLocalSizes(self) -> tuple[int, ...]:
        """Return local elementwise sizes in each dimension.

        Not collective.

        The returned value is calculated excluding ghost points.

        See Also
        --------
        getGlobalSizes, petsc.DMStagGetLocalSizes

        """
        cdef PetscInt dim=0, m=PETSC_DECIDE, n=PETSC_DECIDE, p=PETSC_DECIDE
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetLocalSizes(self.dm, &m, &n, &p))
        return toStagDims(dim, m, n, p)

    def getGlobalSizes(self) -> tuple[int, ...]:
        """Return global element counts in each dimension.

        Not collective.

        See Also
        --------
        getLocalSizes, petsc.DMStagGetGlobalSizes

        """
        cdef PetscInt dim=0, m=PETSC_DECIDE, n=PETSC_DECIDE, p=PETSC_DECIDE
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetGlobalSizes(self.dm, &m, &n, &p))
        return toStagDims(dim, m, n, p)

    def getProcSizes(self) -> tuple[int, ...]:
        """Return number of processes in each dimension.

        Not collective.

        See Also
        --------
        petsc.DMStagGetNumRanks

        """
        cdef PetscInt dim=0, m=PETSC_DECIDE, n=PETSC_DECIDE, p=PETSC_DECIDE
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetNumRanks(self.dm, &m, &n, &p))
        return toStagDims(dim, m, n, p)

    def getStencilType(self) -> str:
        """Return elementwise ghost/halo stencil type.

        Not collective.

        See Also
        --------
        setStencilType, petsc.DMStagGetStencilType

        """
        cdef PetscDMStagStencilType stype = DMSTAG_STENCIL_BOX
        CHKERR(DMStagGetStencilType(self.dm, &stype))
        return toStagStencil(stype)

    def getOwnershipRanges(self) -> tuple[Sequence[int], ...]:
        """Return elements per process in each dimension.

        Not collective.

        See Also
        --------
        setOwnershipRanges, petsc.DMStagGetOwnershipRanges

        """
        cdef PetscInt dim=0, m=0, n=0, p=0
        cdef const PetscInt *lx = NULL, *ly = NULL, *lz = NULL
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetNumRanks(self.dm, &m, &n, &p))
        CHKERR(DMStagGetOwnershipRanges(self.dm, &lx, &ly, &lz))
        return toStagOwnershipRanges(dim, m, n, p, lx, ly, lz)

    def getBoundaryTypes(self) -> tuple[str, ...]:
        """Return boundary types in each dimension.

        Not collective.

        See Also
        --------
        setBoundaryTypes, petsc.DMStagGetBoundaryTypes

        """
        cdef PetscInt dim=0
        cdef PetscDMBoundaryType btx = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType bty = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType btz = DM_BOUNDARY_NONE
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetBoundaryTypes(self.dm, &btx, &bty, &btz))
        return toStagBoundaryTypes(dim, btx, bty, btz)

    def getIsFirstRank(self) -> tuple[int, ...]:
        """Return whether this process is first in each dimension in the process grid.

        Not collective.

        See Also
        --------
        petsc.DMStagGetIsFirstRank

        """
        cdef PetscBool rank0=PETSC_FALSE, rank1=PETSC_FALSE, rank2=PETSC_FALSE
        cdef PetscInt dim=0
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetIsFirstRank(self.dm, &rank0, &rank1, &rank2))
        return toStagDims(dim, rank0, rank1, rank2)

    def getIsLastRank(self) -> tuple[int, ...]:
        """Return whether this process is last in each dimension in the process grid.

        Not collective.

        See Also
        --------
        petsc.DMStagGetIsLastRank

        """
        cdef PetscBool rank0=PETSC_FALSE, rank1=PETSC_FALSE, rank2=PETSC_FALSE
        cdef PetscInt dim=0
        CHKERR(DMGetDimension(self.dm, &dim))
        CHKERR(DMStagGetIsLastRank(self.dm, &rank0, &rank1, &rank2))
        return toStagDims(dim, rank0, rank1, rank2)

    # Coordinate-related functions

    def setUniformCoordinatesExplicit(
        self,
        xmin: float = 0,
        xmax: float = 1,
        ymin: float = 0,
        ymax: float = 1,
        zmin: float = 0,
        zmax: float = 1) -> None:
        """Set coordinates to be a uniform grid, storing all values.

        Collective.

        Parameters
        ----------
        xmin
            The minimum global coordinate value in the x dimension.
        xmax
            The maximum global coordinate value in the x dimension.
        ymin
            The minimum global coordinate value in the y dimension.
        ymax
            The maximum global coordinate value in the y dimension.
        zmin
            The minimum global coordinate value in the z dimension.
        zmax
            The maximum global coordinate value in the z dimension.

        See Also
        --------
        setUniformCoordinatesProduct, setUniformCoordinates
        petsc.DMStagSetUniformCoordinatesExplicit

        """
        cdef PetscReal _xmin = asReal(xmin), _xmax = asReal(xmax)
        cdef PetscReal _ymin = asReal(ymin), _ymax = asReal(ymax)
        cdef PetscReal _zmin = asReal(zmin), _zmax = asReal(zmax)
        CHKERR(DMStagSetUniformCoordinatesExplicit(self.dm, _xmin, _xmax, _ymin, _ymax, _zmin, _zmax))

    def setUniformCoordinatesProduct(
        self,
        xmin: float = 0,
        xmax: float = 1,
        ymin: float = 0,
        ymax: float = 1,
        zmin: float = 0,
        zmax: float = 1) -> None:
        """Create uniform coordinates, as a product of 1D arrays.

        Collective.

        The per-dimension 1-dimensional `DMStag` objects that comprise the
        product always have active 0-cells (vertices, element boundaries) and
        1-cells (element centers).

        Parameters
        ----------
        xmin
            The minimum global coordinate value in the x dimension.
        xmax
            The maximum global coordinate value in the x dimension.
        ymin
            The minimum global coordinate value in the y dimension.
        ymax
            The maximum global coordinate value in the y dimension.
        zmin
            The minimum global coordinate value in the z dimension.
        zmax
            The maximum global coordinate value in the z dimension.

        See Also
        --------
        setUniformCoordinatesExplicit, setUniformCoordinates
        petsc.DMStagSetUniformCoordinatesProduct

        """
        cdef PetscReal _xmin = asReal(xmin), _xmax = asReal(xmax)
        cdef PetscReal _ymin = asReal(ymin), _ymax = asReal(ymax)
        cdef PetscReal _zmin = asReal(zmin), _zmax = asReal(zmax)
        CHKERR(DMStagSetUniformCoordinatesProduct(self.dm, _xmin, _xmax, _ymin, _ymax, _zmin, _zmax))

    def setUniformCoordinates(
        self,
        xmin: float = 0,
        xmax: float = 1,
        ymin: float = 0,
        ymax: float = 1,
        zmin: float = 0,
        zmax: float = 1) -> None:
        """Set the coordinates to be a uniform grid..

        Collective.

        Local coordinates are populated, linearly extrapolated to ghost cells,
        including those outside the physical domain. This is also done in case
        of periodic boundaries, meaning that the same global point may have
        different coordinates in different local representations, which are
        equivalent assuming a periodicity implied by the arguments to this
        function, i.e., two points are equivalent if their difference is a
        multiple of ``xmax-xmin`` in the x dimension, ``ymax-ymin`` in the y
        dimension, and ``zmax-zmin`` in the z dimension.

        Parameters
        ----------
        xmin
            The minimum global coordinate value in the x dimension.
        xmax
            The maximum global coordinate value in the x dimension.
        ymin
            The minimum global coordinate value in the y dimension.
        ymax
            The maximum global coordinate value in the y dimension.
        zmin
            The minimum global coordinate value in the z dimension.
        zmax
            The maximum global coordinate value in the z dimension.

        See Also
        --------
        setUniformCoordinatesExplicit, setUniformCoordinatesProduct
        petsc.DMStagSetUniformCoordinates

        """
        cdef PetscReal _xmin = asReal(xmin), _xmax = asReal(xmax)
        cdef PetscReal _ymin = asReal(ymin), _ymax = asReal(ymax)
        cdef PetscReal _zmin = asReal(zmin), _zmax = asReal(zmax)
        CHKERR(DMStagSetUniformCoordinates(self.dm, _xmin, _xmax, _ymin, _ymax, _zmin, _zmax))

    def setCoordinateDMType(self, dmtype: DM.Type) -> None:
        """Set the type to store coordinates.

        Logically collective.

        Parameters
        ----------
        dmtype
            The type to store coordinates.

        See Also
        --------
        petsc.DMStagSetCoordinateDMType

        """
        cdef PetscDMType cval = NULL
        dmtype = str2bytes(dmtype, &cval)
        CHKERR(DMStagSetCoordinateDMType(self.dm, cval))

    # Location slot related functions

    def getLocationSlot(self, loc: StencilLocation, c: int) -> int:
        """Return index to use in accessing raw local arrays.

        Not collective.

        Parameters
        ----------
        loc
            Location relative to an element.
        c
            Component.

        See Also
        --------
        petsc.DMStagGetLocationSlot

        """
        cdef PetscInt slot=0
        cdef PetscInt comp=asInt(c)
        cdef PetscDMStagStencilLocation sloc = asStagStencilLocation(loc)
        CHKERR(DMStagGetLocationSlot(self.dm, sloc, comp, &slot))
        return toInt(slot)

    def getProductCoordinateLocationSlot(self, loc: StencilLocation) -> None:
        """Return slot for use with local product coordinate arrays.

        Not collective.

        Parameters
        ----------
        loc
            The grid location.

        See Also
        --------
        petsc.DMStagGetProductCoordinateLocationSlot

        """
        cdef PetscInt slot=0
        cdef PetscDMStagStencilLocation sloc = asStagStencilLocation(loc)
        CHKERR(DMStagGetProductCoordinateLocationSlot(self.dm, sloc, &slot))
        return toInt(slot)

    def getLocationDof(self, loc: StencilLocation) -> int:
        """Return number of DOFs associated with a given point on the grid.

        Not collective.

        Parameters
        ----------
        loc
            The grid point.

        See Also
        --------
        petsc.DMStagGetLocationDOF

        """
        cdef PetscInt dof=0
        cdef PetscDMStagStencilLocation sloc = asStagStencilLocation(loc)
        CHKERR(DMStagGetLocationDOF(self.dm, sloc, &dof))
        return toInt(dof)

    # Random other functions

    def migrateVec(self, Vec vec, DM dmTo, Vec vecTo) -> None:
        """Transfer a vector between two ``DMStag`` objects.

        Collective.

        Currently only implemented to migrate global vectors to global vectors.

        Parameters
        ----------
        vec
            The source vector.
        dmTo
            The compatible destination object.
        vecTo
            The destination vector.

        See Also
        --------
        petsc.DMStagMigrateVec

        """
        CHKERR(DMStagMigrateVec(self.dm, vec.vec, dmTo.dm, vecTo.vec))

    def createCompatibleDMStag(self, dofs: tuple[int, ...]) -> DM:
        """Create a compatible ``DMStag`` with different DOFs/stratum.

        Collective.

        Parameters
        ----------
        dofs
            The number of DOFs on the strata in the new `DMStag`.

        See Also
        --------
        petsc.DMStagCreateCompatibleDMStag

        """
        cdef tuple gdofs = tuple(dofs)
        cdef PetscInt dof0=1, dof1=0, dof2=0, dof3=0
        asDofs(gdofs, &dof0, &dof1, &dof2, &dof3)
        cdef PetscDM newda = NULL
        CHKERR(DMStagCreateCompatibleDMStag(self.dm, dof0, dof1, dof2, dof3, &newda))
        cdef DM newdm = type(self)()
        CHKERR(PetscCLEAR(newdm.obj)); newdm.dm = newda
        return newdm

    def VecSplitToDMDA(
        self,
        Vec vec,
        loc: StencilLocation,
        c: int) -> tuple[DMDA, Vec]:
        """Return ``DMDA``, ``Vec`` from a subgrid of a ``DMStag``, its ``Vec``.

        Collective.

        If a ``c`` value of ``-k`` is provided, the first ``k`` DOFs for that
        position are extracted, padding with zero values if needed. If a
        non-negative value is provided, a single DOF is extracted.

        Parameters
        ----------
        vec
            The ``Vec`` object.
        loc
            Which subgrid to extract.
        c
            Which component to extract.

        See Also
        --------
        petsc.DMStagVecSplitToDMDA

        """
        cdef PetscInt pc = asInt(c)
        cdef PetscDMStagStencilLocation sloc = asStagStencilLocation(loc)
        cdef PetscDM pda = NULL
        cdef PetscVec pdavec = NULL
        CHKERR(DMStagVecSplitToDMDA(self.dm, vec.vec, sloc, pc, &pda, &pdavec))
        cdef DM da = DMDA()
        CHKERR(PetscCLEAR(da.obj)); da.dm = pda
        cdef Vec davec = Vec()
        CHKERR(PetscCLEAR(davec.obj)); davec.vec = pdavec
        return (da, davec)

    def getVecArray(self, Vec vec) -> None:
        """Not implemented."""
        raise NotImplementedError('getVecArray for DMStag not yet implemented in petsc4py')

    def get1dCoordinatecArrays(self) -> None:
        """Not implemented."""
        raise NotImplementedError('get1dCoordinatecArrays for DMStag not yet implemented in petsc4py')

    property dim:
        """The dimension."""
        def __get__(self) -> int:
            return self.getDim()

    property dofs:
        """The number of DOFs associated with each stratum of the grid."""
        def __get__(self) -> tuple[int, ...]:
            return self.getDof()

    property entries_per_element:
        """The number of entries per element in the local representation."""
        def __get__(self) -> int:
            return self.getEntriesPerElement()

    property global_sizes:
        """Global element counts in each dimension."""
        def __get__(self) -> tuple[int, ...]:
            return self.getGlobalSizes()

    property local_sizes:
        """Local elementwise sizes in each dimension."""
        def __get__(self) -> tuple[int, ...]:
            return self.getLocalSizes()

    property proc_sizes:
        """The number of processes in each dimension in the global decomposition."""
        def __get__(self) -> tuple[int, ...]:
            return self.getProcSizes()

    property boundary_types:
        """Boundary types in each dimension."""
        def __get__(self) -> tuple[str, ...]:
            return self.getBoundaryTypes()

    property stencil_type:
        """Stencil type."""
        def __get__(self) -> str:
            return self.getStencilType()

    property stencil_width:
        """Elementwise stencil width."""
        def __get__(self) -> int:
            return self.getStencilWidth()

    property corners:
        """The lower left corner and size of local region in each dimension."""
        def __get__(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
            return self.getCorners()

    property ghost_corners:
        """The lower left corner and size of local region in each dimension."""
        def __get__(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
            return self.getGhostCorners()


# --------------------------------------------------------------------

del DMStagStencilType
del DMStagStencilLocation

# --------------------------------------------------------------------



try: range = xrange
except NameError: pass

import sys, petsc4py
petsc4py.init(sys.argv)

from petsc4py import PETSc

m, n  = 16, 32

A = PETSc.Mat().create(PETSc.COMM_WORLD)
A.setSizes([m*n, m*n])
A.setFromOptions()
A.setUp()
Istart, Iend = A.getOwnershipRange()
for I in range(Istart, Iend):
    A[I,I] = 4
    i = I//n
    if i>0  : J = I-n; A[I,J] = -1
    if i<m-1: J = I+n; A[I,J] = -1
    j = I-i*n
    if j>0  : J = I-1; A[I,J] = -1
    if j<n-1: J = I+1; A[I,J] = -1
A.assemblyBegin()
A.assemblyEnd()

x, y = A.createVecs()
x.set(1)
A.mult(x,y)

# save
viewer = PETSc.Viewer().createBinary('matrix-A.dat', 'w')
viewer(A)
viewer = PETSc.Viewer().createBinary('vector-x.dat', 'w')
viewer(x)
viewer = PETSc.Viewer().createBinary('vector-y.dat', 'w')
viewer(y)

# load
viewer = PETSc.Viewer().createBinary('matrix-A.dat', 'r')
B = PETSc.Mat().load(viewer)
viewer = PETSc.Viewer().createBinary('vector-x.dat', 'r')
u = PETSc.Vec().load(viewer)
viewer = PETSc.Viewer().createBinary('vector-y.dat', 'r')
v = PETSc.Vec().load(viewer)

# check
assert B.equal(A)
assert x.equal(u)
assert y.equal(v)



cdef extern from * nogil:

    ctypedef double PetscLogDouble
    ctypedef struct PetscEventPerfInfo:
        int count
        PetscLogDouble flops, time
        PetscLogDouble numMessages
        PetscLogDouble messageLength
        PetscLogDouble numReductions

    PetscErrorCode PetscLogDefaultBegin()
    PetscErrorCode PetscLogView(PetscViewer)
    PetscErrorCode PetscLogIsActive(PetscBool*)

    PetscErrorCode PetscLogFlops(PetscLogDouble)
    PetscErrorCode PetscGetFlops(PetscLogDouble*)
    PetscErrorCode PetscGetCPUTime(PetscLogDouble*)
    PetscErrorCode PetscMallocGetCurrentUsage(PetscLogDouble*)
    PetscErrorCode PetscMemoryGetCurrentUsage(PetscLogDouble*)

    PetscErrorCode PetscTime(PetscLogDouble*)
    PetscErrorCode PetscTimeSubtract(PetscLogDouble*)
    PetscErrorCode PetscTimeAdd(PetscLogDouble*)

    ctypedef int PetscLogStage
    PetscErrorCode PetscLogStageRegister(char[], PetscLogStage*)
    PetscErrorCode PetscLogStagePush(PetscLogStage)
    PetscErrorCode PetscLogStagePop()
    PetscErrorCode PetscLogStageSetActive(PetscLogStage, PetscBool)
    PetscErrorCode PetscLogStageGetActive(PetscLogStage, PetscBool*)
    PetscErrorCode PetscLogStageSetVisible(PetscLogStage, PetscBool)
    PetscErrorCode PetscLogStageGetVisible(PetscLogStage, PetscBool*)
    PetscErrorCode PetscLogStageGetId(char[], PetscLogStage*)

    ctypedef int PetscLogClass "PetscClassId"
    PetscErrorCode PetscLogClassRegister"PetscClassIdRegister"(char[], PetscLogClass*)
    PetscErrorCode PetscLogClassActivate"PetscLogEventActivateClass"(PetscLogClass)
    PetscErrorCode PetscLogClassDeactivate"PetscLogEventDeactivateClass"(PetscLogClass)

    ctypedef int PetscLogEvent
    PetscErrorCode PetscLogEventRegister(char[], PetscLogClass, PetscLogEvent*)
    PetscErrorCode PetscLogEventBegin(PetscLogEvent, PetscObject, PetscObject, PetscObject, PetscObject)
    PetscErrorCode PetscLogEventEnd(PetscLogEvent, PetscObject, PetscObject, PetscObject, PetscObject)

    PetscErrorCode PetscLogEventActivate(PetscLogEvent)
    PetscErrorCode PetscLogEventDeactivate(PetscLogEvent)
    PetscErrorCode PetscLogEventSetActiveAll(PetscLogEvent, PetscBool)
    PetscErrorCode PetscLogEventGetPerfInfo(PetscLogStage, PetscLogEvent, PetscEventPerfInfo*)

cdef extern from * nogil: # custom.h
    PetscErrorCode PetscLogStageFindId(char[], PetscLogStage*)
    PetscErrorCode PetscLogClassFindId(char[], PetscLogClass*)
    PetscErrorCode PetscLogEventFindId(char[], PetscLogEvent*)
    PetscErrorCode PetscLogStageFindName(PetscLogStage, char*[])
    PetscErrorCode PetscLogClassFindName(PetscLogClass, char*[])
    PetscErrorCode PetscLogEventFindName(PetscLogEvent, char*[])


cdef inline int event_args2objs(object args, PetscObject o[4]) except -1:
    o[0] = o[1] = o[2] = o[3] = NULL
    cdef Py_ssize_t i=0, n = len(args)
    cdef Object tmp = None
    if n > 4: n = 4
    for 0 <= i < n:
        tmp = args[i]
        if tmp is not None:
            o[i] = tmp.obj[0]
    return 0



cdef extern from * nogil:

    ctypedef const char* PetscAOType "AOType"
    PetscAOType AOBASIC
    PetscAOType AOADVANCED
    PetscAOType AOMAPPING
    PetscAOType AOMEMORYSCALABLE

    PetscErrorCode AOView(PetscAO, PetscViewer)
    PetscErrorCode AODestroy(PetscAO*)
    PetscErrorCode AOCreateBasic(MPI_Comm, PetscInt, const PetscInt[], const PetscInt[], PetscAO*)
    PetscErrorCode AOCreateBasicIS(PetscIS, PetscIS, PetscAO*)
    PetscErrorCode AOCreateMemoryScalable(MPI_Comm, PetscInt, const PetscInt[], const PetscInt[], PetscAO*)
    PetscErrorCode AOCreateMemoryScalableIS(PetscIS, PetscIS, PetscAO*)
    PetscErrorCode AOCreateMapping(MPI_Comm, PetscInt, const PetscInt[], const PetscInt[], PetscAO*)
    PetscErrorCode AOCreateMappingIS(PetscIS, PetscIS, PetscAO*)
    PetscErrorCode AOGetType(PetscAO, PetscAOType*)

    PetscErrorCode AOApplicationToPetsc(PetscAO, PetscInt, PetscInt[])
    PetscErrorCode AOApplicationToPetscIS(PetscAO, PetscIS)
    PetscErrorCode AOPetscToApplication(PetscAO, PetscInt, PetscInt[])
    PetscErrorCode AOPetscToApplicationIS(PetscAO, PetscIS)



import os
import sys
import inspect
import textwrap
from sphinx.util import logging

logger = logging.getLogger(__name__)


def is_cyfunction(obj):
    return type(obj).__name__ == 'cython_function_or_method'


def is_function(obj):
    return inspect.isbuiltin(obj) or is_cyfunction(obj) or type(obj) is type(ord)


def is_method(obj):
    return (
        inspect.ismethoddescriptor(obj)
        or inspect.ismethod(obj)
        or is_cyfunction(obj)
        or type(obj)
        in (
            type(str.index),
            type(str.__add__),
            type(str.__new__),
        )
    )


def is_classmethod(obj):
    return inspect.isbuiltin(obj) or type(obj).__name__ in (
        'classmethod',
        'classmethod_descriptor',
    )


def is_staticmethod(obj):
    return type(obj).__name__ in ('staticmethod',)


def is_constant(obj):
    return isinstance(obj, (int, float, str, dict))


def is_datadescr(obj):
    return inspect.isdatadescriptor(obj) and not hasattr(obj, 'fget')


def is_property(obj):
    return inspect.isdatadescriptor(obj) and hasattr(obj, 'fget')


def is_class(obj):
    return inspect.isclass(obj) or type(obj) is type(int)


def is_hidden(obj):
    return obj.__qualname__.startswith('_')


class Lines(list):
    INDENT = ' ' * 4
    level = 0

    @property
    def add(self):
        return self

    @add.setter
    def add(self, lines):
        if lines is None:
            return
        if isinstance(lines, str):
            lines = textwrap.dedent(lines).strip().split('\n')
        indent = self.INDENT * self.level
        for line in lines:
            self.append(indent + line)


def signature(obj, fail=True):
    doc = obj.__doc__
    if not doc:
        if fail and not is_hidden(obj):
            logger.warning(f'Missing signature for {obj}')
        doc = f'{obj.__name__}: Any'
    sig = doc.partition('\n')[0].split('.', 1)[-1]
    return sig or None


def docstring(obj, fail=True):
    doc = obj.__doc__
    if not doc:
        if fail and not is_hidden(obj):
            logger.warning(f'Missing docstring for {obj}')
        doc = ''
    link = None
    sig = None
    cl = is_class(obj)
    if cl:
        doc = doc.strip()
    else:
        sig, _, doc = doc.partition('\n')
        doc, _, link = doc.rpartition('\n')

    summary, _, docbody = doc.partition('\n')
    summary = summary.strip()
    docbody = textwrap.dedent(docbody).strip()

    # raise warning if docstring is not provided for a method
    if not summary and not is_function(obj) and is_method(obj):
        logger.warning(f'docstring: Missing summary for {obj}')

    # warnings for docstrings that are not compliant
    if len(summary) > 79:
        logger.warning(f'Summary for {obj} too long.')
    if docbody:
        if not summary.endswith('.'):
            logger.warning(f'Summary for {obj} does not end with period.')
        # FIXME
        lines = docbody.split('\n')
        for i, line in enumerate(lines):
            if len(line) > 79:
                logger.warning(f'Line {i} for documentation of {obj} too long.')
        if not cl:
            init = (
                'Collective.',
                'Not collective.',
                'Logically collective.',
                'Neighborwise collective.',
                'Collective the first time it is called.',
            )
            if lines[0] not in init:
                logger.warning(f'Unexpected collectiveness for {sig}\nFound {lines[0]}')

    if link:
        linktxt, _, link = link.rpartition(' ')
        linkloc = link.replace(':', '#L')
        # FIXME do we want to use a special section?
        # section = f'References\n----------`'
        section = '\n'
        linkbody = f':sources:`{linktxt} {link} <{linkloc}>`'
        linkbody = f'{section}\n{linkbody}'
        if docbody:
            docbody = f'{docbody}\n\n{linkbody}'
        else:
            docbody = linkbody

    if docbody:
        doc = f'"""{summary}\n\n{docbody}\n\n"""'
    else:
        doc = f'"""{summary}"""'
    return textwrap.indent(doc, Lines.INDENT)


def visit_data(constant):
    name, value = constant
    typename = type(value).__name__
    kind = 'Constant' if isinstance(value, int) else 'Object'
    init = f"_def({typename}, '{name}')"
    doc = f'#: {kind} ``{name}`` of type :class:`{typename}`'
    return f'{name}: {typename} = {init}  {doc}\n'


def visit_function(function):
    sig = signature(function)
    doc = docstring(function)
    body = Lines.INDENT + '...'
    return f'def {sig}:\n{doc}\n{body}\n'


def visit_method(method):
    sig = signature(method)
    doc = docstring(method)
    body = Lines.INDENT + '...'
    return f'def {sig}:\n{doc}\n{body}\n'


def visit_datadescr(datadescr, name=None):
    sig = signature(datadescr)
    doc = docstring(datadescr)
    name = sig.partition(':')[0].strip() or datadescr.__name__
    rtype = sig.partition(':')[2].strip() or 'Any'
    sig = f'{name}(self) -> {rtype}'
    body = Lines.INDENT + '...'
    return f'@property\ndef {sig}:\n{doc}\n{body}\n'


def visit_property(prop, name=None):
    sig = signature(prop.fget)
    name = name or prop.fget.__name__
    rtype = sig.rsplit('->', 1)[-1].strip()
    sig = f'{name}(self) -> {rtype}'
    doc = f'"""{prop.__doc__}"""'
    doc = textwrap.indent(doc, Lines.INDENT)
    body = Lines.INDENT + '...'
    return f'@property\ndef {sig}:\n{doc}\n{body}\n'


def visit_constructor(cls, name='__init__', args=None):
    init = name == '__init__'
    argname = cls.__mro__[-2].__name__.lower()
    argtype = cls.__name__
    initarg = args or f'{argname}: Optional[{argtype}] = None'
    selfarg = 'self' if init else 'cls'
    rettype = 'None' if init else argtype
    arglist = f'{selfarg}, {initarg}'
    sig = f'{name}({arglist}) -> {rettype}'
    ret = '...' if init else 'return super().__new__(cls)'
    body = Lines.INDENT + ret
    return f'def {sig}:\n{body}'


def visit_class(cls, outer=None, done=None):
    skip = {
        '__doc__',
        '__dict__',
        '__module__',
        '__weakref__',
        '__pyx_vtable__',
        '__lt__',
        '__le__',
        '__ge__',
        '__gt__',
        '__enum2str',  # FIXME refactor implementation
        '_traceback_',  # FIXME maybe refactor?
    }
    special = {
        '__len__': '__len__(self) -> int',
        '__bool__': '__bool__(self) -> bool',
        '__hash__': '__hash__(self) -> int',
        '__int__': '__int__(self) -> int',
        '__index__': '__int__(self) -> int',
        '__str__': '__str__(self) -> str',
        '__repr__': '__repr__(self) -> str',
        '__eq__': '__eq__(self, other: object) -> bool',
        '__ne__': '__ne__(self, other: object) -> bool',
    }

    qualname = cls.__name__
    cls_name = cls.__name__
    if outer is not None and cls_name.startswith(outer):
        cls_name = cls_name[len(outer) :]
        qualname = f'{outer}.{cls_name}'

    override = OVERRIDE.get(qualname, {})
    done = set() if done is None else done
    lines = Lines()

    base = cls.__base__
    if base is object:
        lines.add = f'class {cls_name}:'
    else:
        lines.add = f'class {cls_name}({base.__name__}):'
    lines.level += 1

    lines.add = docstring(cls)

    for name in ('__new__', '__init__', '__hash__'):
        if name in cls.__dict__:
            done.add(name)

    dct = cls.__dict__
    keys = list(dct.keys())

    def dunder(name):
        return name.startswith('__') and name.endswith('__')

    def members(seq):
        for name in seq:
            if name in skip:
                continue
            if name in done:
                continue
            if dunder(name):
                if name not in special and name not in override:
                    done.add(name)
                    continue
            yield name

    for name in members(keys):
        attr = getattr(cls, name)
        if is_class(attr):
            done.add(name)
            lines.add = visit_class(attr, outer=cls_name)
            continue

    for name in members(keys):
        if name in override:
            done.add(name)
            lines.add = override[name]
            continue

        if name in special:
            done.add(name)
            sig = special[name]
            lines.add = f'def {sig}: ...'
            continue

        attr = getattr(cls, name)

        if is_method(attr):
            done.add(name)
            if name == attr.__name__:
                obj = dct[name]
                if is_classmethod(obj):
                    lines.add = '@classmethod'
                elif is_staticmethod(obj):
                    lines.add = '@staticmethod'
                lines.add = visit_method(attr)
            continue

        if is_datadescr(attr):
            done.add(name)
            lines.add = visit_datadescr(attr)
            continue

        if is_property(attr):
            done.add(name)
            lines.add = visit_property(attr, name)
            continue

        if is_constant(attr):
            done.add(name)
            lines.add = visit_data((name, attr))
            continue

    leftovers = [name for name in keys if name not in done and name not in skip]
    if leftovers:
        raise RuntimeError(f'leftovers: {leftovers}')

    lines.level -= 1
    return lines


def visit_module(module, done=None):
    skip = {
        '__doc__',
        '__name__',
        '__loader__',
        '__spec__',
        '__file__',
        '__package__',
        '__builtins__',
        '__pyx_capi__',
        '__pyx_unpickle_Enum',  # FIXME review
    }

    done = set() if done is None else done
    lines = Lines()

    keys = list(module.__dict__.keys())
    keys.sort(key=lambda name: name.startswith('_'))

    constants = [
        (name, getattr(module, name))
        for name in keys
        if all(
            (
                name not in done and name not in skip,
                is_constant(getattr(module, name)),
            )
        )
    ]
    for _, value in constants:
        cls = type(value)
        name = cls.__name__
        if name in done or name in skip:
            continue
        if cls.__module__ == module.__name__:
            done.add(name)
            lines.add = visit_class(cls)
            lines.add = ''
    for attr in constants:
        name, value = attr
        done.add(name)
        if name in OVERRIDE:
            lines.add = OVERRIDE[name]
        else:
            lines.add = visit_data((name, value))
    if constants:
        lines.add = ''

    for name in keys:
        if name in done or name in skip:
            continue
        value = getattr(module, name)

        if is_class(value):
            done.add(name)
            if value.__name__ != name:
                continue
            if value.__module__ != module.__name__:
                continue
            lines.add = visit_class(value)
            lines.add = ''
            instances = [
                (k, getattr(module, k))
                for k in keys
                if all(
                    (
                        k not in done and k not in skip,
                        type(getattr(module, k)) is value,
                    )
                )
            ]
            for attrname, attrvalue in instances:
                done.add(attrname)
                lines.add = visit_data((attrname, attrvalue))
            if instances:
                lines.add = ''
            continue

        if is_function(value):
            done.add(name)
            if name == value.__name__:
                lines.add = visit_function(value)
            else:
                lines.add = f'{name} = {value.__name__}'
            continue

    lines.add = ''
    for name in keys:
        if name in done or name in skip:
            continue
        value = getattr(module, name)
        done.add(name)
        if name in OVERRIDE:
            lines.add = OVERRIDE[name]
        else:
            lines.add = visit_data((name, value))

    leftovers = [name for name in keys if name not in done and name not in skip]
    if leftovers:
        raise RuntimeError(f'leftovers: {leftovers}')
    return lines


IMPORTS = """
from __future__ import annotations
import sys
from typing import (
    Any,
    Union,
    Literal,
    Optional,
    NoReturn,
    Final,
)
from typing import (
    Callable,
    Hashable,
    Iterable,
    Iterator,
    Sequence,
    Mapping,
)
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import numpy
from numpy import dtype, ndarray
from mpi4py.MPI import (
    Intracomm,
    Datatype,
    Op,
)

class _dtype:
    def __init__(self, name):
        self.name = name
    def __repr__(self):
        return self.name

IntType: dtype = _dtype('IntType')
RealType: dtype =  _dtype('RealType')
ComplexType: dtype = _dtype('ComplexType')
ScalarType: dtype = _dtype('ScalarType')
"""

HELPERS = """
class _Int(int): pass
class _Str(str): pass
class _Float(float): pass
class _Dict(dict): pass

def _repr(obj):
    try:
        return obj._name
    except AttributeError:
        return super(obj).__repr__()

def _def(cls, name):
    if cls is int:
       cls = _Int
    if cls is str:
       cls = _Str
    if cls is float:
       cls = _Float
    if cls is dict:
       cls = _Dict

    obj = cls()
    obj._name = name
    if '__repr__' not in cls.__dict__:
        cls.__repr__ = _repr
    return obj
"""

OVERRIDE = {}

TYPING = """
from .typing import *
"""


def visit_petsc4py_PETSc(done=None):
    from petsc4py import PETSc

    lines = Lines()
    lines.add = f'"""{PETSc.__doc__}"""'
    lines.add = IMPORTS
    lines.add = ''
    lines.add = HELPERS
    lines.add = ''
    lines.add = visit_module(PETSc)
    lines.add = ''
    lines.add = TYPING
    return lines


def generate(filename):
    dirname = os.path.dirname(filename)
    os.makedirs(dirname, exist_ok=True)
    with open(filename, 'w') as f:
        for line in visit_petsc4py_PETSc():
            print(line, file=f)


def load_module(filename, name=None):
    if name is None:
        name, _ = os.path.splitext(os.path.basename(filename))
    module = type(sys)(name)
    module.__file__ = filename
    module.__package__ = name.rsplit('.', 1)[0]
    old = replace_module(module)
    with open(filename) as f:
        exec(f.read(), module.__dict__)  # noqa: S102
    restore_module(old)
    return module


_sys_modules = {}


def replace_module(module):
    name = module.__name__
    if name in _sys_modules:
        raise RuntimeError(f'{name} in modules')
    _sys_modules[name] = sys.modules[name]
    sys.modules[name] = module
    return _sys_modules[name]


def restore_module(module):
    name = module.__name__
    if name not in _sys_modules:
        raise RuntimeError(f'{name} not in modules')
    sys.modules[name] = _sys_modules[name]
    del _sys_modules[name]


def annotate(dest, source):
    try:
        dest.__annotations__ = source.__annotations__
    except AttributeError:
        pass
    if isinstance(dest, type):
        for name in dest.__dict__.keys():
            if hasattr(source, name):
                obj = getattr(dest, name)
                annotate(obj, getattr(source, name))
    if isinstance(dest, type(sys)):
        for name in dir(dest):
            if hasattr(source, name):
                obj = getattr(dest, name)
                mod = getattr(obj, '__module__', None)
                if dest.__name__ == mod:
                    annotate(obj, getattr(source, name))
        for name in dir(source):
            if not hasattr(dest, name):
                setattr(dest, name, getattr(source, name))


OUTDIR = 'reference'

if __name__ == '__main__':
    generate(os.path.join(OUTDIR, 'petsc4py.PETSc.py'))



from petsc4py import PETSc
import unittest
import numpy as np

# --------------------------------------------------------------------


class BaseTestVec:
    COMM = None
    TYPE = None

    def setUp(self):
        v = PETSc.Vec()
        v.create(self.COMM)
        v.setSizes(100)
        v.setType(self.TYPE)
        self.vec = v

    def tearDown(self):
        self.vec.destroy()
        self.vec = None
        PETSc.garbage_cleanup()

    def testDuplicate(self):
        self.vec.set(1)
        vec = self.vec.duplicate()
        self.assertFalse(self.vec.equal(vec))
        self.assertEqual(self.vec.sizes, vec.sizes)
        del vec

    def testCopy(self):
        self.vec.set(1)
        vec = self.vec.duplicate()
        self.vec.copy(vec)
        self.assertTrue(self.vec.equal(vec))
        del vec

    def testDot(self):
        self.vec.set(1)
        d = self.vec.dot(self.vec)
        self.assertAlmostEqual(abs(d), self.vec.getSize())
        self.vec.dotBegin(self.vec)
        d = self.vec.dotEnd(self.vec)
        self.assertAlmostEqual(abs(d), self.vec.getSize())

    def testNorm(self):
        from math import sqrt

        self.vec.set(1)
        n1 = self.vec.norm(PETSc.NormType.NORM_1)
        n2 = self.vec.norm(PETSc.NormType.NORM_2)
        ni = self.vec.norm(PETSc.NormType.NORM_INFINITY)
        self.assertAlmostEqual(n1, self.vec.getSize())
        self.assertAlmostEqual(n2, sqrt(self.vec.getSize()))
        self.assertAlmostEqual(n2, self.vec.norm())
        self.assertAlmostEqual(ni, 1.0)
        self.vec.normBegin(PETSc.NormType.NORM_1)
        nn1 = self.vec.normEnd(PETSc.NormType.NORM_1)
        self.assertAlmostEqual(nn1, n1)
        self.vec.normBegin()
        nn2 = self.vec.normEnd()
        self.assertAlmostEqual(nn2, n2)
        self.vec.normBegin(PETSc.NormType.NORM_INFINITY)
        nni = self.vec.normEnd(PETSc.NormType.NORM_INFINITY)
        self.assertAlmostEqual(nni, ni)

    def testNormalize(self):
        from math import sqrt

        self.vec.set(1)
        n2 = self.vec.normalize()
        self.assertAlmostEqual(n2, sqrt(self.vec.getSize()))
        self.assertAlmostEqual(1, self.vec.norm())

    def testSumMinMax(self):
        self.vec.set(1)
        self.assertEqual(self.vec.sum(), self.vec.getSize())
        self.vec.set(-7)
        self.assertEqual(self.vec.min()[1], -7)
        self.vec.set(10)
        self.assertEqual(self.vec.max()[1], 10)

    def testSwap(self):
        v1 = self.vec
        v2 = v1.duplicate()
        v1.set(1)
        v2.set(2)
        v1.swap(v2)
        idx, _ = self.vec.getOwnershipRange()
        self.assertEqual(v1[idx], 2)
        self.assertEqual(v2[idx], 1)

    def testBsize(self):
        self.vec.setBlockSize(1)
        self.assertEqual(self.vec.getBlockSize(), 1)
        self.vec.setBlockSize(1)

    def testGetSetVals(self):
        start, end = self.vec.getOwnershipRange()
        self.vec[start] = -7
        self.vec[end - 1] = -7
        self.vec.assemble()
        self.assertEqual(self.vec[start], -7)
        self.assertEqual(self.vec[end - 1], -7)
        for i in range(start, end):
            self.vec[i] = i
        self.vec.assemble()
        values = [self.vec[i] for i in range(start, end)]
        self.assertEqual(values, list(range(start, end)))
        sz = self.vec.getSize()
        self.assertEqual(self.vec.sum(), (sz - 1) / 2.0 * sz)

    def testGetSetValsBlocked(self):
        return
        lsize, gsize = self.vec.getSizes()
        start, end = self.vec.getOwnershipRange()
        bsizes = list(range(1, lsize + 1))
        nblocks = list(range(1, lsize + 1))
        compat = [
            (bs, nb)
            for bs in bsizes
            if not (gsize % bs or lsize % bs)
            for nb in nblocks
            if bs * nb <= lsize
        ]
        for bsize, nblock in compat:
            self.vec.setBlockSize(bsize)
            bindex = [start // bsize + i for i in range(nblock)]
            bvalue = [float(i) for i in range(nblock * bsize)]
            self.vec.setValuesBlocked(bindex, bvalue)
            self.vec.assemble()
            index = [start + i for i in range(nblock * bsize)]
            value = self.vec.getValues(index)
            self.assertEqual(bvalue, list(value))

    def testGetSetArray(self):
        self.vec.set(1)
        arr0 = self.vec.getArray().copy()
        self.assertEqual(arr0.sum(), self.vec.getLocalSize())
        arr0 = self.vec.getArray().copy()
        self.vec.setRandom()
        arr1 = self.vec.getArray().copy()
        self.vec.setArray(arr1)
        arr1 = self.vec.getArray().copy()
        arr2 = self.vec.getArray().copy()
        self.assertTrue((arr1 == arr2).all())

        refs = self.vec.getRefCount()
        arr3 = np.asarray(self.vec)
        self.assertEqual(self.vec.getRefCount(), refs + 1)
        self.assertTrue((arr1 == arr3).all())
        arr3[:] = 0
        self.assertAlmostEqual(abs(self.vec.sum()), 0)
        self.assertEqual(self.vec.max()[1], 0)
        self.assertEqual(self.vec.min()[1], 0)
        self.vec.set(1)
        self.assertAlmostEqual(abs(arr3.sum()), self.vec.getLocalSize())
        self.assertEqual(arr3.min(), 1)
        self.assertEqual(arr3.max(), 1)
        del arr3
        self.assertEqual(self.vec.getRefCount(), refs)

    def testPlaceArray(self):
        self.vec.set(1)
        array = self.vec.getArray().copy()
        self.vec.placeArray(array)
        array[:] = 2
        self.assertAlmostEqual(abs(self.vec.sum()), 2 * self.vec.getSize())
        self.vec.resetArray()
        self.assertAlmostEqual(abs(self.vec.sum()), self.vec.getSize())

    def testLocalVector(self):
        rank = self.vec.getComm().Get_rank()
        self.vec.getArray()[:] = rank + 1
        ln = self.vec.getLocalSize()
        lvec = self.vec.createLocalVector()
        self.vec.getLocalVector(lvec)
        self.assertEqual(abs(lvec.sum()), (rank + 1) * ln)
        self.vec.restoreLocalVector(lvec)
        self.vec.getLocalVector(lvec, readonly=True)
        self.assertEqual(abs(lvec.sum()), (rank + 1) * ln)
        self.vec.restoreLocalVector(lvec, readonly=True)
        lvec.destroy()

    def testSetOption(self):
        opt1 = PETSc.Vec.Option.IGNORE_OFF_PROC_ENTRIES
        opt2 = PETSc.Vec.Option.IGNORE_NEGATIVE_INDICES
        for opt in [opt1, opt2] * 2:
            for flag in [True, False] * 2:
                self.vec.setOption(opt, flag)

    def testGetSetItem(self):
        v = self.vec
        w = v.duplicate()
        #
        v[...] = 7
        self.assertEqual(v.max()[1], 7)
        self.assertEqual(v.min()[1], 7)
        #
        v.setRandom()
        w[...] = v
        self.assertTrue(w.equal(v))
        #
        v.setRandom()
        w[...] = v.getArray()
        self.assertTrue(w.equal(v))
        #
        s, e = v.getOwnershipRange()
        v.setRandom()
        w[s:e] = v.getArray().copy()
        w.assemble()
        self.assertTrue(w.equal(v))
        w1, v1 = w[s], v[s]
        w2, v2 = w[e - 1], v[e - 1]
        self.assertEqual(w1, v1)
        self.assertEqual(w2, v2)

    def testMAXPY(self):
        y = self.vec
        y.set(1)
        x = [y.copy() for _ in range(3)]
        a = [1] * len(x)
        y.maxpy(a, x)
        z = y.duplicate()
        z.set(len(x) + 1)
        self.assertTrue(y.equal(z))

    def testBinOp(self):
        x = self.vec
        x.set(1)
        n = x.getSize()
        y = 2 + 2 * x + 1 - x * 3 - 1
        self.assertEqual(y.min()[1], 1)
        self.assertEqual(y.max()[1], 1)
        z = (4 * x) / (2 * y)
        self.assertEqual(z.min()[1], 2)
        self.assertEqual(z.max()[1], 2)
        z = z / 2
        self.assertEqual(z.min()[1], 1)
        self.assertEqual(z.max()[1], 1)
        s = (+x) @ (-y)
        self.assertEqual(s, -n)
        #
        M, N = n, 2 * n
        A = PETSc.Mat().createDense((M, N), comm=self.COMM)
        A.setUp()
        rs, re = A.getOwnershipRange()
        cs, ce = A.getOwnershipRangeColumn()
        a, b = 3, 5
        for i in range(rs, re):
            for j in range(N):
                A[i, j] = a * i + b * j
        A.assemble()
        y = x @ A
        self.assertEqual(y.getSize(), N)
        for i in range(cs, ce):
            self.assertEqual(y[i], a * M * (M - 1) / 2 + b * i * M)
        y.set(1)
        z = A @ y
        self.assertEqual(z.getSize(), M)
        for i in range(rs, re):
            self.assertEqual(z[i], b * N * (N - 1) / 2 + a * i * N)

    def testConcatenate(self):
        x = self.vec
        y = x.duplicate()
        x.set(1)
        y.set(2)
        z, index_ises = PETSc.Vec.concatenate([x, y])
        self.assertEqual(z.getLocalSize(), x.getLocalSize() + y.getLocalSize())
        self.assertEqual(z.min()[1], x.min()[1])
        self.assertEqual(z.max()[1], y.max()[1])
        np.allclose(z.getArray(), np.concatenate([x.getArray(), y.getArray()]))
        np.allclose(z.getArray()[0 : x.getLocalSize()], x.getArray())
        np.allclose(z.getArray()[x.getLocalSize() :], y.getArray())

    def testMean(self):
        x = self.vec
        x.getArray()[:] = self.COMM.rank + 1
        self.assertAlmostEqual(x.mean(), np.mean(np.arange(1, self.COMM.size + 1)))


# --------------------------------------------------------------------


class TestVecSeq(BaseTestVec, unittest.TestCase):
    COMM = PETSc.COMM_SELF
    TYPE = PETSc.Vec.Type.SEQ


class TestVecMPI(BaseTestVec, unittest.TestCase):
    COMM = PETSc.COMM_WORLD
    TYPE = PETSc.Vec.Type.MPI


class TestVecShared(BaseTestVec, unittest.TestCase):
    if PETSc.COMM_WORLD.getSize() == 1:
        TYPE = PETSc.Vec.Type.SHARED
    else:
        TYPE = PETSc.Vec.Type.MPI
    COMM = PETSc.COMM_WORLD


# class TestVecSieve(BaseTestVec, unittest.TestCase):
#    CLASS = PETSc.VecSieve
#    TARGS = ([],)

# class TestVecGhost(BaseTestVec, unittest.TestCase):
#    CLASS = PETSc.VecGhost
#    TARGS = ([],)

# --------------------------------------------------------------------


class TestVecWithArray(unittest.TestCase):
    def testCreateSeq(self):
        a = np.zeros(5, dtype=PETSc.ScalarType)

        v1 = PETSc.Vec().createWithArray(a, comm=PETSc.COMM_SELF)
        v2 = PETSc.Vec().createWithArray(a, size=5, comm=PETSc.COMM_SELF)
        v3 = PETSc.Vec().createWithArray(a, size=3, comm=PETSc.COMM_SELF)

        self.assertTrue(v1.size == 5)
        self.assertTrue(v2.size == 5)
        self.assertTrue(v3.size == 3)

        a1 = v1.getDict()['__array__']
        self.assertTrue(a is a1)
        a2 = v2.getDict()['__array__']
        self.assertTrue(a is a2)
        a3 = v3.getDict()['__array__']
        self.assertTrue(a is a3)

    def testCreateMPI(self):
        a = np.zeros(5, dtype=PETSc.ScalarType)

        v1 = PETSc.Vec().createWithArray(a, comm=PETSc.COMM_WORLD)
        v2 = PETSc.Vec().createWithArray(a, size=(5, None), comm=PETSc.COMM_WORLD)
        v3 = PETSc.Vec().createWithArray(a, size=(3, None), comm=PETSc.COMM_WORLD)

        self.assertTrue(v1.local_size == 5)
        self.assertTrue(v2.local_size == 5)
        self.assertTrue(v3.local_size == 3)

        a1 = v1.getDict()['__array__']
        self.assertTrue(a is a1)
        a2 = v2.getDict()['__array__']
        self.assertTrue(a is a2)
        a3 = v3.getDict()['__array__']
        self.assertTrue(a is a3)

    def testSetMPIGhost(self):
        v = PETSc.Vec().create()
        v.setType(PETSc.Vec.Type.MPI)
        v.setSizes((5, None))
        ghosts = [i % v.size for i in range(v.owner_range[1], v.owner_range[1] + 3)]
        v.setMPIGhost(ghosts)
        v.setArray(np.array(range(*v.owner_range), dtype=PETSc.ScalarType))
        v.ghostUpdate()
        with v.localForm() as loc:
            self.assertTrue((loc[0 : v.local_size] == range(*v.owner_range)).all())
            self.assertTrue((loc[v.local_size :] == ghosts).all())

    def testGetGhostIS(self):
        v = PETSc.Vec().create()
        v.setType(PETSc.Vec.Type.MPI)
        v.setSizes((5, None))
        ghosts = [i % v.size for i in range(v.owner_range[1], v.owner_range[1] + 3)]
        v.setMPIGhost(ghosts)
        v.setArray(np.array(range(*v.owner_range), dtype=PETSc.ScalarType))
        v.ghostUpdate()
        self.assertTrue((v.getGhostIS().getIndices() == ghosts).all())


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()

# --------------------------------------------------------------------



try:
    execfile
except NameError:
    def execfile(file, globals=globals(), locals=locals()):
        fh = open(file, "r")
        try: exec(fh.read()+"\n", globals, locals)
        finally: fh.close()

import petsc4py, sys
petsc4py.init(sys.argv)

from petsc4py import PETSc

execfile('petsc-mat.py')
execfile('petsc-ksp.py')

OptDB = PETSc.Options()

if OptDB.getBool('plot', True):
    da = PETSc.DMDA().create([m,n])
    u = da.createGlobalVec()
    x.copy(u)
    draw = PETSc.Viewer.DRAW()
    OptDB['draw_pause'] = 1
    draw(u)

if OptDB.getBool('plot_mpl', False):
    try:
        from matplotlib import pylab
    except ImportError:
        print("matplotlib not available")
    else:
        from numpy import mgrid
        X, Y =  mgrid[0:1:1j*m,0:1:1j*n]
        Z = x[...].reshape(m,n)
        pylab.figure()
        pylab.contourf(X,Y,Z)
        pylab.plot(X.ravel(),Y.ravel(),'.k')
        pylab.axis('equal')
        pylab.colorbar()
        pylab.show()



#!/usr/bin/env python3
import sys, petsc4py
petsc4py.init(sys.argv)
from petsc4py import PETSc
import numpy as np

try:
    from matplotlib import pylab
except ImportError:
    pylab = None

# this user class is an application
# context for the nonlinear problem
# at hand; it contains some parameters
# and knows how to compute residuals

class AppCtx:

    def __init__(self, nx, ny, nz):
        self.n = np.array([nx, ny, nz], dtype='i')
        self.h = np.array([1.0/(n-1) for n in self.n], dtype='d')
        from App import formFunction
        from App import formInitial
        self._formFunction = formFunction
        self._formInitial = formInitial

    def formInitial(self, t, X):
        xx = X.getArray(readonly=0).reshape(self.n, order='f')
        self._formInitial(self.h, t, xx)

    def formFunction(self, ts, t, X, Xdot, F):
        n = self.n
        h = self.h
        x = X.getArray(readonly=1).reshape(n, order='f')
        xdot = Xdot.getArray(readonly=1).reshape(n, order='f')
        f = F[...].reshape(n, order='f')
        self._formFunction(h, t, x, xdot, f)

    def plot(self, t, x):
        nx, ny, nz = self.n
        from numpy import mgrid
        #
        U = x.getArray(readonly=1).reshape(nx,ny,nz, order='f')
        #
        X, Y =  mgrid[0:1:1j*nx,0:1:1j*ny]
        Z = U[:,:,nz//2]
        pylab.figure(0)
        pylab.contourf(X,Y,Z)
        pylab.colorbar()
        pylab.plot(X.ravel(),Y.ravel(),'.k')
        pylab.title('z=0.50')
        pylab.xlabel('x')
        pylab.ylabel('y')
        pylab.axis('equal')
        #
        X, Y =  mgrid[0:1:1j*nx,0:1:1j*nz]
        Z = U[:,ny//4,:]
        pylab.figure(1)
        pylab.contourf(X,Y,Z)
        pylab.colorbar()
        pylab.plot(X.ravel(),Y.ravel(),'.k')
        pylab.title('y=0.25')
        pylab.xlabel('x')
        pylab.ylabel('z')
        pylab.axis('equal')
        #
        X, Y =  mgrid[0:1:1j*ny,0:1:1j*nz]
        Z = U[nx//2,:,:]
        pylab.figure(2)
        pylab.contourf(X,Y,Z)
        pylab.colorbar()
        pylab.plot(X.ravel(),Y.ravel(),'.k')
        pylab.title('x=0.50')
        pylab.xlabel('y')
        pylab.ylabel('z')
        pylab.axis('equal')


def run_test(nx,ny,nz,samples,plot=False):
    ts  = PETSc.TS().create()
    ts.setType('theta')
    ts.setTheta(1.0)
    ts.setTimeStep(0.01)
    ts.setTime(0.0)
    ts.setMaxTime(1.0)
    ts.setMaxSteps(10)
    eft = PETSc.TS.ExactFinalTime.STEPOVER
    ts.setExactFinalTime(eft)

    x = PETSc.Vec().createSeq(nx*ny*nz)
    ts.setSolution(x)
    app = AppCtx(nx, ny, nz)
    f = PETSc.Vec().createSeq(nx*ny*nz)
    ts.setIFunction(app.formFunction, f)
    ts.snes.setUseMF(1)
    ts.snes.ksp.setType('cg')
    ts.setFromOptions()
    ts.setUp()

    wt = 1e300
    for i in range(samples):
        app.formInitial(0, x)
        t1 = PETSc.Log.getTime()
        ts.solve(x)
        t2 = PETSc.Log.getTime()
        wt = min(wt,t2-t1)
        
    if plot and pylab: 
        app.plot(ts.time, x)

    return wt

OptDB = PETSc.Options()


start = OptDB.getInt('start', 12)
step = OptDB.getInt('step', 4)
stop = OptDB.getInt('stop', start)
samples = OptDB.getInt('samples', 1)

plot = OptDB.getBool('plot', False)
if plot and not pylab:
    PETSc.Sys.Print("matplotlib not available")

for n in range(start, stop+step, step):
    nx = ny = nz = n+1
    wt = run_test(nx,ny,nz,samples,plot)
    PETSc.Sys.Print("Grid %3d x %3d x %3d -> %f  seconds (%2d samples)" 
                    % (nx,ny,nz,wt,samples))

if plot and pylab:
    pylab.show()




'''
Ex2 from PETSc example files implemented for PETSc4py.
https://petsc.org/release/src/ksp/ksp/tutorials/ex2.c.html
By: Miguel Arriaga


Solves a linear system in parallel with KSP.
Input parameters include:
    -view_exact_sol   : write exact solution vector to stdout
    -m <mesh_x>       : number of mesh points in x-direction
    -n <mesh_n>       : number of mesh points in y-direction


Vec            x,b,u;    # approx solution, RHS, exact solution
Mat            A;        # linear system matrix
KSP            ksp;      # linear solver context
PetscReal      norm;     # norm of solution error
'''
import sys
import petsc4py
petsc4py.init(sys.argv)
from petsc4py import PETSc

import numpy as np

comm = PETSc.COMM_WORLD
size = comm.getSize()
rank = comm.getRank()

OptDB = PETSc.Options()
m  = OptDB.getInt('m', 8)
n  = OptDB.getInt('n', 7)

''' - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        Compute the matrix and right-hand-side vector that define
        the linear system, Ax = b.
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - '''
'''
    Create parallel matrix, specifying only its global dimensions.
    When using MatCreate(), the matrix format can be specified at
    runtime. Also, the parallel partitioning of the matrix is
    determined by PETSc at runtime.

    Performance tuning note:  For problems of substantial size,
    preallocation of matrix memory is crucial for attaining good
    performance. See the matrix chapter of the users manual for details.
'''

A = PETSc.Mat().create(comm=comm)
A.setSizes((m*n,m*n))
A.setFromOptions()
A.setPreallocationNNZ((5,5))

'''
    Currently, all PETSc parallel matrix formats are partitioned by
    contiguous chunks of rows across the processors.  Determine which
    rows of the matrix are locally owned.
'''
Istart,Iend = A.getOwnershipRange()

'''
    Set matrix elements for the 2-D, five-point stencil in parallel.
    - Each processor needs to insert only elements that it owns
    locally (but any non-local elements will be sent to the
    appropriate processor during matrix assembly).
    - Always specify global rows and columns of matrix entries.

    Note: this uses the less common natural ordering that orders first
    all the unknowns for x = h then for x = 2h etc; Hence you see J = Ii +- n
    instead of J = I +- m as you might expect. The more standard ordering
    would first do all variables for y = h, then y = 2h etc.
'''

for Ii in range(Istart,Iend):
    v = -1.0
    i = int(Ii/n)
    j = int(Ii - i*n)

    if (i>0):
        J = Ii - n
        A.setValues(Ii,J,v,addv=True)
    if (i<m-1):
        J = Ii + n
        A.setValues(Ii,J,v,addv=True)
    if (j>0):
        J = Ii - 1
        A.setValues(Ii,J,v,addv=True)
    if (j<n-1):
        J = Ii + 1
        A.setValues(Ii,J,v,addv=True)

    v = 4.0
    A.setValues(Ii,Ii,v,addv=True)

'''
    Assemble matrix, using the 2-step process:
    MatAssemblyBegin(), MatAssemblyEnd()
    Computations can be done while messages are in transition
    by placing code between these two statements.
'''

A.assemblyBegin(A.AssemblyType.FINAL)
A.assemblyEnd(A.AssemblyType.FINAL)
''' A is symmetric. Set symmetric flag to enable ICC/Cholesky preconditioner '''

A.setOption(A.Option.SYMMETRIC,True)

'''
    Create parallel vectors.
    - We form 1 vector from scratch and then duplicate as needed.
    - When using VecCreate(), VecSetSizes and VecSetFromOptions()
    in this example, we specify only the
    vector's global dimension; the parallel partitioning is determined
    at runtime.
    - When solving a linear system, the vectors and matrices MUST
    be partitioned accordingly.  PETSc automatically generates
    appropriately partitioned matrices and vectors when MatCreate()
    and VecCreate() are used with the same communicator.
    - The user can alternatively specify the local vector and matrix
    dimensions when more sophisticated partitioning is needed
    (replacing the PETSC_DECIDE argument in the VecSetSizes() statement
    below).
'''

u = PETSc.Vec().create(comm=comm)
u.setSizes(m*n)
u.setFromOptions()

b = u.duplicate()
x = b.duplicate()

'''
    Set exact solution; then compute right-hand-side vector.
    By default we use an exact solution of a vector with all
    elements of 1.0;
'''
u.set(1.0)
b = A(u)

'''
    View the exact solution vector if desired
'''
flg = OptDB.getBool('view_exact_sol', False)
if flg:
    u.view()

''' - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            Create the linear solver and set various options
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - '''
ksp = PETSc.KSP().create(comm=comm)

'''
    Set operators. Here the matrix that defines the linear system
    also serves as the matrix from which the preconditioner is constructed.
'''
ksp.setOperators(A,A)

'''
    Set linear solver defaults for this problem (optional).
    - By extracting the KSP and PC contexts from the KSP context,
    we can then directly call any KSP and PC routines to set
    various options.
    - The following two statements are optional; all of these
    parameters could alternatively be specified at runtime via
    KSPSetFromOptions().  All of these defaults can be
    overridden at runtime, as indicated below.
'''
rtol=1.e-2/((m+1)*(n+1))
ksp.setTolerances(rtol=rtol,atol=1.e-50)

'''
Set runtime options, e.g.,
    -ksp_type <type> -pc_type <type> -ksp_monitor -ksp_rtol <rtol>
These options will override those specified above as long as
KSPSetFromOptions() is called _after_ any other customization
routines.
'''
ksp.setFromOptions()

''' - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                    Solve the linear system
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - '''

ksp.solve(b,x)

''' - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                    Check the solution and clean up
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - '''
x = x - u # x.axpy(-1.0,u)
norm = x.norm(PETSc.NormType.NORM_2)
its = ksp.getIterationNumber()

'''
    Print convergence information.  PetscPrintf() produces a single
    print statement from all processes that share a communicator.
    An alternative is PetscFPrintf(), which prints to a file.
'''
if norm > rtol*10:
    PETSc.Sys.Print("Norm of error {}, Iterations {}".format(norm,its),comm=comm)
else:
    if size==1:
        PETSc.Sys.Print("- Serial OK",comm=comm)
    else:
        PETSc.Sys.Print("- Parallel OK",comm=comm)



cdef extern from * nogil:

    ctypedef const char* PetscMatPartitioningType "MatPartitioningType"
    PetscMatPartitioningType MATPARTITIONINGCURRENT
    PetscMatPartitioningType MATPARTITIONINGAVERAGE
    PetscMatPartitioningType MATPARTITIONINGSQUARE
    PetscMatPartitioningType MATPARTITIONINGPARMETIS
    PetscMatPartitioningType MATPARTITIONINGCHACO
    PetscMatPartitioningType MATPARTITIONINGPARTY
    PetscMatPartitioningType MATPARTITIONINGPTSCOTCH
    PetscMatPartitioningType MATPARTITIONINGHIERARCH

    PetscErrorCode MatPartitioningCreate(MPI_Comm, PetscMatPartitioning*)
    PetscErrorCode MatPartitioningDestroy(PetscMatPartitioning*)
    PetscErrorCode MatPartitioningView(PetscMatPartitioning, PetscViewer)

    PetscErrorCode MatPartitioningSetType(PetscMatPartitioning, PetscMatPartitioningType)
    PetscErrorCode MatPartitioningGetType(PetscMatPartitioning, PetscMatPartitioningType*)
    PetscErrorCode MatPartitioningSetFromOptions(PetscMatPartitioning)

    PetscErrorCode MatPartitioningSetAdjacency(PetscMatPartitioning, PetscMat)
    PetscErrorCode MatPartitioningApply(PetscMatPartitioning, PetscIS*)



import sys, petsc4py
petsc4py.init(sys.argv)

from petsc4py import PETSc
import Bratu3D as Bratu3D

class App(object):

    def __init__(self, da, lambda_):
        assert da.getDim() == 3
        self.da = da
        self.params = Bratu3D.Params()
        self.params.lambda_ = lambda_

    def formInitGuess(self, X):
        X.zeroEntries() # just in case
        Bratu3D.FormInitGuess(self.da, X, self.params)

    def formFunction(self, snes, X, F):
        F.zeroEntries() # just in case
        Bratu3D.FormFunction(self.da, X, F, self.params)

    def formJacobian(self, snes, X, J, P):
        P.zeroEntries() # just in case
        Bratu3D.FormJacobian(self.da, X, P, self.params)
        if J != P: J.assemble() # matrix-free operator
        return PETSc.Mat.Structure.SAME_NONZERO_PATTERN


OptDB = PETSc.Options()

N = OptDB.getInt('N', 16)
lambda_ = OptDB.getReal('lambda', 6.0)
do_plot = OptDB.getBool('plot', False)

da = PETSc.DMDA().create([N, N, N], stencil_width=1)
app = App(da, lambda_)

snes = PETSc.SNES().create()
F = da.createGlobalVec()
snes.setFunction(app.formFunction, F)
J = da.createMat()
snes.setJacobian(app.formJacobian, J)

snes.setFromOptions()

X = da.createGlobalVec()
app.formInitGuess(X)
snes.solve(None, X)

U = da.createNaturalVec()
da.globalToNatural(X, U)


def plot(da, U):
    scatter, U0 = PETSc.Scatter.toZero(U)
    scatter.scatter(U, U0, False, PETSc.Scatter.Mode.FORWARD)
    rank = PETSc.COMM_WORLD.getRank()
    if rank == 0:
        solution = U0[...].reshape(da.sizes, order='f').copy()
        try:
            from matplotlib import pyplot
            pyplot.contourf(solution[:, :, N//2])
            pyplot.axis('equal')
            pyplot.show()
        except:
            pass
    PETSc.COMM_WORLD.barrier()
    scatter.destroy()
    U0.destroy()


if do_plot: plot(da, U)


U.destroy()
X.destroy()
F.destroy()
J.destroy()
da.destroy()
snes.destroy()



from petsc4py import PETSc
import os
import unittest
import gc
import weakref

# --------------------------------------------------------------------

## gc.set_debug((gc.DEBUG_STATS |
##               gc.DEBUG_LEAK) &
##              ~gc.DEBUG_SAVEALL)

# --------------------------------------------------------------------

PYSABI = os.path.basename(PETSc.__file__).split(".")[1] == "abi3"


class BaseTestGC:
    def setUp(self):
        self.obj = self.CLASS().create(comm=PETSc.COMM_SELF)

    def tearDown(self):
        wref = self.make_weakref()
        self.assertTrue(wref() is self.obj)
        self.obj = None
        gc.collect()
        if PYSABI:
            # garbage collection is not supported
            # the weak reference is still alive
            self.assertTrue(wref() is not None)
            self.assertEqual(wref().getRefCount(), 1)
            wref().destroy()
        else:
            # the weak reference is gone
            self.assertTrue(wref() is None)
        del wref
        PETSc.garbage_cleanup()

    def make_weakref(self):
        return weakref.ref(self.obj)

    def testCycleInSelf(self):
        self.obj.setAttr('myself', self.obj)

    def testCycleInMethod(self):
        self.obj.setAttr('mymeth', self.obj.view)

    def testCycleInInstance(self):
        class A:
            pass

        a = A()
        a.obj = self.obj
        self.obj.setAttr('myinst', a)

    def testCycleInAllWays(self):
        self.testCycleInSelf()
        self.testCycleInMethod()
        self.testCycleInInstance()


# --------------------------------------------------------------------


class TestGCVec(BaseTestGC, unittest.TestCase):
    CLASS = PETSc.Vec


class TestGCVecSubType(TestGCVec):
    CLASS = type('_Vec', (PETSc.Vec,), {})


class TestGCMat(BaseTestGC, unittest.TestCase):
    CLASS = PETSc.Mat


class TestGCMatSubType(TestGCMat):
    CLASS = type('_Mat', (PETSc.Mat,), {})


class TestGCPC(BaseTestGC, unittest.TestCase):
    CLASS = PETSc.PC


class TestGCPCSubType(TestGCPC):
    CLASS = type('_PC', (PETSc.PC,), {})


class TestGCKSP(BaseTestGC, unittest.TestCase):
    CLASS = PETSc.KSP


class TestGCKSPSubType(TestGCKSP):
    CLASS = type('_KSP', (PETSc.KSP,), {})


class TestGCSNES(BaseTestGC, unittest.TestCase):
    CLASS = PETSc.SNES

    def testCycleInAppCtx(self):
        self.obj.setAppCtx(self.obj)


class TestGCSNESSubType(TestGCSNES):
    CLASS = type('_SNES', (PETSc.SNES,), {})


class TestGCTS(BaseTestGC, unittest.TestCase):
    CLASS = PETSc.TS

    def testCycleInAppCtx(self):
        self.obj.setAppCtx(self.obj)


class TestGCTSSubType(TestGCTS):
    CLASS = type('_TS', (PETSc.TS,), {})

    def testCycleInAppCtx(self):
        self.obj.setAppCtx(self.obj)


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()



import sys,petsc4py
petsc4py.init(sys.argv)
from petsc4py import PETSc
import numpy as np

OptDB = PETSc.Options()

dim = OptDB.getInt('dim', 2)
plex = PETSc.DMPlex().createBoxMesh([4]*dim, simplex=True)
plex.distribute()
plex.view()

# Create two metric tensor fields corresponding to uniform mesh sizes of 0.1 and 0.2
metric1 = plex.metricCreateUniform(100.0)
metric2 = plex.metricCreateUniform(25.0)

# The metrics can be combined using intersection, the result of which corresponds to
# the maximum ellipsoid at each point
metric = plex.metricCreate()
plex.metricIntersection2(metric1, metric2, metric)
metric1.axpy(-1, metric)
assert np.isclose(metric1.norm(), 0.0)

# Call adapt routine - boundary label None by default
newplex = plex.adaptMetric(metric)
newplex.view()

# Write to VTK file
viewer = PETSc.Viewer().createVTK('base_mesh.vtk', 'w')
viewer(plex)
viewer = PETSc.Viewer().createVTK('isotropic_mesh.vtk', 'w')
viewer(newplex)



from petsc4py import PETSc
import unittest

class BaseTestPC:
    KSP_TYPE = None
    PC_TYPE = None
    def setUp(self):
        ksp = PETSc.KSP()
        ksp.create(PETSc.COMM_SELF)
        pc = ksp.getPC()
        if self.KSP_TYPE:
            ksp.setType(self.KSP_TYPE)
        if self.PC_TYPE:
            pc.setType(self.PC_TYPE)
        self.ksp = ksp
        self.pc = pc

    def tearDown(self):
        self.ksp = None
        self.pc = None
        PETSc.garbage_cleanup()

class TestFIELDSPLITPC(BaseTestPC, unittest.TestCase):
    PC_TYPE = PETSc.PC.Type.FIELDSPLIT

    def testISoperations(self):
        test_index = [0,1,2]
        pc = self.pc
        is_u = PETSc.IS().createGeneral(test_index, comm=PETSc.COMM_SELF)
        pc.setFieldSplitIS(("u", is_u))

        self.assertTrue((pc.getFieldSplitSubIS("u").getIndices() == test_index).all())
        is_u = None


if __name__ == '__main__':
    unittest.main()



from petsc4py.PETSc import KSP
from petsc4py.PETSc import PC
from petsc4py.PETSc import Mat
from petsc4py.PETSc import Vec
from petsc4py.PETSc import Viewer


# A template class with the Python methods supported by PCPYTHON


class PCPythonProtocol:
    def apply(self, pc: PC, b: Vec, x: Vec) -> None:
        """Apply the preconditioner on vector b, return in x."""
        ...

    def applySymmetricLeft(self, pc: PC, b: Vec, x: Vec) -> None:
        """Apply the symmetric left part of the preconditioner on vector b, return in x."""
        ...

    def applySymmetricRight(self, pc: PC, b: Vec, x: Vec) -> None:
        """Apply the symmetric right part of the preconditioner on vector b, return in x."""
        ...

    def applyTranspose(self, pc: PC, b: Vec, x: Vec) -> None:
        """Apply the transposed preconditioner on vector b, return in x."""
        ...

    def applyMat(self, pc: PC, B: Mat, X: Mat) -> None:
        """Apply the preconditioner on a block of right-hand sides B, return in X."""
        ...

    def preSolve(self, pc: PC, ksp: KSP, b: Vec, x: Vec) -> None:
        """Callback called at the beginning of a Krylov method.

        This method is allowed to modify the right-hand side b and the initial guess x.

        """
        ...

    def postSolve(self, pc: PC, ksp: KSP, b: Vec, x: Vec) -> None:
        """Callback called at the end of a Krylov method.

        This method is allowed to modify the right-hand side b and the solution x.

        """

    def view(self, pc: PC, viewer: Viewer) -> None:
        """View the preconditioner."""
        ...

    def setFromOptions(self, pc: PC) -> None:
        """Process command line for customization."""
        ...

    def setUp(self, pc: PC) -> None:
        """Perform the required setup."""
        ...

    def reset(self, pc: PC) -> None:
        """Reset the preconditioner."""
        ...



# --------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import petsc4py

    petsc4py.init(sys.argv + ['-log_view'])

# --------------------------------------------------------------------

from petsc4py import PETSc
import unittest

# --------------------------------------------------------------------


class TestLog(unittest.TestCase):
    def setUp(self):
        # PETSc.Log.begin()
        # register stages
        self.stage1 = PETSc.Log.Stage('Stage 1')
        self.stage2 = PETSc.Log.Stage('Stage 2')
        # register classes
        self.klassA = PETSc.Log.Class('Class A')
        self.klassB = PETSc.Log.Class('Class B')
        # register events
        self.event1 = PETSc.Log.Event('Event 1')  # no class
        self.event2 = PETSc.Log.Event('Event 2')  # no class
        self.eventA = PETSc.Log.Event('Event A', self.klassA)
        self.eventB = PETSc.Log.Event('Event B', self.klassB)

    def testGetName(self):
        self.assertEqual(self.klassA.name, 'Class A')
        self.assertEqual(self.klassB.name, 'Class B')
        self.assertEqual(self.event1.name, 'Event 1')
        self.assertEqual(self.event2.name, 'Event 2')
        self.assertEqual(self.eventA.name, 'Event A')
        self.assertEqual(self.eventB.name, 'Event B')
        self.assertEqual(self.stage1.name, 'Stage 1')
        self.assertEqual(self.stage2.name, 'Stage 2')

    def testLogBeginEnd(self):
        # -----
        self._run_events()  # in main stage
        self._run_stages()  # in user stages
        # -----
        for event in self._get_events():
            event.deactivate()
            event.setActive(False)
            event.active = False
        self._run_events()  # should not be logged
        for event in self._get_events():
            event.activate()
            event.setActive(True)
            event.active = True
        # -----
        for klass in self._get_classes():
            klass.deactivate()
            klass.setActive(False)
            klass.active = False
        self._run_events()  # A and B should not be logged
        for klass in self._get_classes():
            klass.activate()
            klass.setActive(True)
            klass.active = True
        # -----
        for stage in self._get_stages():
            active = stage.getActive()
            self.assertTrue(active)
            self.assertTrue(stage.active)
            stage.setActive(False)
            active = stage.getActive()
            self.assertFalse(active)
            self.assertFalse(stage.active)
        self._run_stages()  # should not be logged
        for stage in self._get_stages():
            stage.setActive(True)
            stage.active = True
            active = stage.getActive()
            self.assertTrue(active)
            self.assertTrue(stage.active)
        # -----
        self._run_events()
        self._run_stages()

    def test_log_decorator(self):
        timed_func = PETSc.Log.EventDecorator()(lambda: None)
        timed_func()

    def _run_stages(self):
        for stage in self._get_stages():
            self._run_events(stage)

    def _get_stages(self):
        return (self.stage1, self.stage2)

    def _get_classes(self):
        return (self.klassA, self.klassB)

    def _get_events(self):
        return (self.event1, self.event2, self.eventA, self.eventB)

    def _run_events(self, stage=None):
        if stage is not None:
            stage.push()
        self._events_begin()
        self._events_end()
        if stage is not None:
            stage.pop()

    def _events_begin(self):
        for event in self._get_events():
            event.begin()

    def _events_end(self):
        for event in reversed(self._get_events()):
            event.end()


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()
